"""Types used in Quokka"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import enum

import quokka
from typing import (
    Any,
    DefaultDict,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

AddressT = int
Index = int
T = TypeVar("T")

LocationValueType = Union[
    Tuple[int, int], Tuple[int, int, int], int  # Structure  # Inst instance
]

RegType = enum.IntEnum

class RegAccessMode(enum.Enum):
    """Register access mode"""

    READ = enum.auto()
    WRITE = enum.auto()
    ANY = enum.auto()

ReferenceTarget = Union[
    "quokka.structure.Structure",
    "quokka.structure.StructureMember",
    "quokka.data.Data",
    "quokka.Instruction",
    "quokka.Chunk",
    Tuple["quokka.Chunk", "quokka.Block", Index],
]


class AddressSize(enum.Enum):
    """Address size"""

    ADDRESS_64 = enum.auto()
    ADDRESS_32 = enum.auto()
    ADDRESS_16 = enum.auto()
    ADDRESS_UNK = enum.auto()

    @staticmethod
    def from_proto(
        address_size: "quokka.pb.Quokka.AddressSizeValue",
    ) -> "AddressSize":
        """Convert the protobuf value into this enumeration"""
        mapping = {
            quokka.pb.Quokka.ADDR_32: AddressSize.ADDRESS_32,
            quokka.pb.Quokka.ADDR_64: AddressSize.ADDRESS_64,
        }

        return mapping.get(address_size, AddressSize.ADDRESS_UNK)


class Endianness(enum.Enum):
    """Endianness of the program

    LE: Little endian (least significant bit first)
    BE: Big endian (most significant bit first)

    TODO:
        See how we can support mixed endianness
    """

    LITTLE_ENDIAN = enum.auto()
    BIG_ENDIAN = enum.auto()
    UNKNOWN = enum.auto()

    @staticmethod
    def from_proto(
        endianness: "quokka.pb.Quokka.Meta.EndianessValue",
    ) -> Endianness:
        """Convert the protobuf value into this enumeration"""
        mapping = {
            quokka.pb.Quokka.Meta.END_BE: Endianness.BIG_ENDIAN,
            quokka.pb.Quokka.Meta.END_LE: Endianness.LITTLE_ENDIAN,
        }

        return mapping.get(endianness, Endianness.UNKNOWN)


class EdgeType(enum.Enum):
    """Edge Type"""

    UNCONDITIONAL = enum.auto()
    TRUE = enum.auto()
    FALSE = enum.auto()
    SWITCH = enum.auto()

    @staticmethod
    def from_proto(
        edge_type: "quokka.pb.Quokka.Edge.EdgeTypeValue",
    ) -> "EdgeType":
        """Convert the protobuf value into this enumeration"""
        mapping = {
            quokka.pb.Quokka.Edge.TYPE_UNCONDITIONAL: EdgeType.UNCONDITIONAL,
            quokka.pb.Quokka.Edge.TYPE_TRUE: EdgeType.TRUE,
            quokka.pb.Quokka.Edge.TYPE_FALSE: EdgeType.FALSE,
            quokka.pb.Quokka.Edge.TYPE_SWITCH: EdgeType.SWITCH,
        }

        edge = mapping.get(edge_type)
        if edge is not None:
            return edge

        raise ValueError("Unable to decode Edge Type")


class FunctionType(enum.Enum):
    """Function Type"""

    NORMAL = enum.auto()
    IMPORTED = enum.auto()
    LIBRARY = enum.auto()
    THUNK = enum.auto()
    EXTERN = enum.auto()
    INVALID = enum.auto()

    @staticmethod
    def from_proto(
        function_type: "quokka.pb.Quokka.Function.FunctionTypeValue",
    ) -> "FunctionType":
        """Convert the protobuf value into this enumeration"""
        mapping = {
            quokka.pb.Quokka.Function.TYPE_NORMAL: FunctionType.NORMAL,
            quokka.pb.Quokka.Function.TYPE_IMPORTED: FunctionType.IMPORTED,
            quokka.pb.Quokka.Function.TYPE_LIBRARY: FunctionType.LIBRARY,
            quokka.pb.Quokka.Function.TYPE_THUNK: FunctionType.THUNK,
        }

        return mapping.get(function_type, FunctionType.INVALID)


class BlockType(enum.Enum):
    """Block Type"""

    NORMAL = enum.auto()
    INDJUMP = enum.auto()
    RET = enum.auto()
    NORET = enum.auto()
    CNDRET = enum.auto()
    ENORET = enum.auto()
    EXTERN = enum.auto()
    ERROR = enum.auto()
    FAKE = enum.auto()

    @staticmethod
    def from_proto(
        block_type: "quokka.pb.Quokka.FunctionChunk.Block.BlockTypeValue",
    ) -> BlockType:
        """Convert the protobuf value into this enumeration"""
        mapping = {
            quokka.pb.Quokka.FunctionChunk.Block.BLOCK_TYPE_NORMAL: BlockType.NORMAL,
            quokka.pb.Quokka.FunctionChunk.Block.BLOCK_TYPE_INDJUMP: BlockType.INDJUMP,
            quokka.pb.Quokka.FunctionChunk.Block.BLOCK_TYPE_RET: BlockType.RET,
            quokka.pb.Quokka.FunctionChunk.Block.BLOCK_TYPE_NORET: BlockType.NORET,
            quokka.pb.Quokka.FunctionChunk.Block.BLOCK_TYPE_CNDRET: BlockType.CNDRET,
            quokka.pb.Quokka.FunctionChunk.Block.BLOCK_TYPE_ENORET: BlockType.ENORET,
            quokka.pb.Quokka.FunctionChunk.Block.BLOCK_TYPE_EXTERN: BlockType.EXTERN,
            quokka.pb.Quokka.FunctionChunk.Block.BLOCK_TYPE_ERROR: BlockType.ERROR,
            quokka.pb.Quokka.FunctionChunk.Block.BLOCK_TYPE_FAKE: BlockType.FAKE,
        }

        return mapping.get(block_type, BlockType.FAKE)


class ReferenceType(enum.Enum):
    """Reference Type"""

    CALL = enum.auto()
    DATA = enum.auto()
    ENUM = enum.auto()
    STRUC = enum.auto()
    UNKNOWN = enum.auto()

    @staticmethod
    def from_proto(
        reference_type: "quokka.pb.Quokka.Reference.ReferenceTypeValue",
    ) -> "ReferenceType":
        """Convert the protobuf value into this enumeration"""
        mapping = {
            quokka.pb.Quokka.Reference.REF_CALL: ReferenceType.CALL,
            quokka.pb.Quokka.Reference.REF_DATA: ReferenceType.DATA,
            quokka.pb.Quokka.Reference.REF_ENUM: ReferenceType.ENUM,
            quokka.pb.Quokka.Reference.REF_STRUC: ReferenceType.STRUC,
        }

        return mapping.get(reference_type, ReferenceType.UNKNOWN)


class DataType(enum.Enum):
    """Data Type"""

    UNKNOWN = enum.auto()
    BYTE = enum.auto()
    WORD = enum.auto()
    DOUBLE_WORD = enum.auto()
    QUAD_WORD = enum.auto()
    OCTO_WORD = enum.auto()
    FLOAT = enum.auto()
    DOUBLE = enum.auto()
    ASCII = enum.auto()
    STRUCT = enum.auto()
    ALIGN = enum.auto()
    POINTER = enum.auto()

    @staticmethod
    def from_proto(data_type: "quokka.pb.Quokka.DataTypeValue") -> "DataType":
        """Convert the protobuf value into this enumeration"""
        mapping = {
            quokka.pb.Quokka.TYPE_B: DataType.BYTE,
            quokka.pb.Quokka.TYPE_W: DataType.WORD,
            quokka.pb.Quokka.TYPE_DW: DataType.DOUBLE_WORD,
            quokka.pb.Quokka.TYPE_QW: DataType.QUAD_WORD,
            quokka.pb.Quokka.TYPE_OW: DataType.QUAD_WORD,
            quokka.pb.Quokka.TYPE_FLOAT: DataType.FLOAT,
            quokka.pb.Quokka.TYPE_DOUBLE: DataType.DOUBLE,
            quokka.pb.Quokka.TYPE_ASCII: DataType.ASCII,
            quokka.pb.Quokka.TYPE_STRUCT: DataType.STRUCT,
            quokka.pb.Quokka.TYPE_ALIGN: DataType.ALIGN,
            quokka.pb.Quokka.TYPE_POINTER: DataType.POINTER,
        }

        return mapping.get(data_type, DataType.UNKNOWN)


class SegmentType(enum.Enum):
    """Segment Type"""

    UNKNOWN = enum.auto()
    CODE = enum.auto()
    DATA = enum.auto()
    BSS = enum.auto()
    NULL = enum.auto()
    EXTERN = enum.auto()
    NORMAL = enum.auto()
    ABSOLUTE_SYMBOLS = enum.auto()

    @staticmethod
    def from_proto(
        segment_type: "quokka.pb.Quokka.Segment.TypeValue",
    ) -> "SegmentType":
        """Convert the protobuf value into this enumeration"""
        mapping = {
            quokka.pb.Quokka.Segment.SEGMENT_CODE: SegmentType.CODE,
            quokka.pb.Quokka.Segment.SEGMENT_DATA: SegmentType.DATA,
            quokka.pb.Quokka.Segment.SEGMENT_BSS: SegmentType.BSS,
            quokka.pb.Quokka.Segment.SEGMENT_NULL: SegmentType.NULL,
            quokka.pb.Quokka.Segment.SEGMENT_NORMAL: SegmentType.NORMAL,
            quokka.pb.Quokka.Segment.SEGMENT_EXTERN: SegmentType.EXTERN,
            quokka.pb.Quokka.Segment.SEGMENT_ABSOLUTE_SYMBOLS: SegmentType.ABSOLUTE_SYMBOLS,
        }

        return mapping.get(segment_type, SegmentType.UNKNOWN)


class StructureType(enum.Enum):
    """Structure Type"""

    STRUCT = enum.auto()
    ENUM = enum.auto()
    UNION = enum.auto()
    UNKNOWN = enum.auto()

    @staticmethod
    def from_proto(
        structure_type: "quokka.pb.Quokka.Structure.StructureTypeValue",
    ) -> "StructureType":
        """Convert the protobuf value into this enumeration"""
        mapping = {
            quokka.pb.Quokka.Structure.TYPE_STRUCT: StructureType.STRUCT,
            quokka.pb.Quokka.Structure.TYPE_ENUM: StructureType.ENUM,
            quokka.pb.Quokka.Structure.TYPE_UNION: StructureType.UNION,
        }

        return mapping.get(structure_type, StructureType.UNKNOWN)


class ExporterMode(enum.IntEnum):
    """Mode type

    The exporter mode controls the type of exported data.
    """

    LIGHT = enum.auto()
    FULL = enum.auto()
    NORMAL = enum.auto()

    @staticmethod
    def from_proto(mode: "quokka.pb.Quokka.ExporterMeta.ModeValue") -> "ExporterMode":
        mapping = {
            quokka.pb.Quokka.ExporterMeta.MODE_LIGHT: ExporterMode.LIGHT,
            quokka.pb.Quokka.ExporterMeta.MODE_FULL: ExporterMode.FULL,
        }

        return mapping.get(mode, ExporterMode.NORMAL)



"""Architecture module"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import collections
from enum import IntEnum, IntFlag

import capstone

from quokka.types import List, RegType
from types import ModuleType


def make_enums(
    capstone_module: ModuleType, items: List[str], blacklist: List[str], flags_enums: List[str]
) -> List:
    """Make enums from capstone module

    Dynamically generate enums from capstone constants

    Arguments:
        capstone_module: Capstone module
        items: Name of enums
        blacklist: Fields to not consider
        flags_enums: List of flag enums

    Returns:
        A list of IntEnum/IntFlag
    """
    data = collections.defaultdict(dict)
    for key, val in getattr(capstone_module, "__dict__").items():
        _, cat, name = key.split("_", maxsplit=2)
        if name not in blacklist:
            name = "_" + name if "0" <= name[0] <= "9" else name
            data[cat][name] = val

    return [
        IntEnum(x, names=data[x], module=__name__)
        if x not in flags_enums
        else IntFlag(x, names=data[x], module=__name__)
        for x in items
        if x
    ]


class ArchEnum(IntEnum):
    """
    Enum identifying various architectures. The integer
    values are based on capstone enum.
    """

    UNKNOWN = capstone.CS_ARCH_ALL
    ARM = capstone.CS_ARCH_ARM
    ARM64 = capstone.CS_ARCH_ARM64
    EVM = capstone.CS_ARCH_EVM
    M680X = capstone.CS_ARCH_M680X
    M68K = capstone.CS_ARCH_M68K
    MIPS = capstone.CS_ARCH_MIPS
    PPC = capstone.CS_ARCH_PPC
    SPARC = capstone.CS_ARCH_SPARC
    SYSZ = capstone.CS_ARCH_SYSZ
    TMS320C64X = capstone.CS_ARCH_TMS320C64X
    X86 = capstone.CS_ARCH_X86
    XCORE = capstone.CS_ARCH_XCORE


class QuokkaArch:
    """Base class for a QuokkaArch"""

    address_size: int
    compared_mnemonics: List[str]
    stack_pointer: RegType
    inst_pointer: RegType


class ArchX86(QuokkaArch):  # type: ignore
    """Arch X86 definition"""

    address_size = 32

    regs, insts, groups, prefixes, optypes, _flags_state = make_enums(
        capstone_module=capstone.x86_const,
        items=["REG", "INS", "GRP", "PREFIX", "OP", "EFLAGS"],
        blacklist=["ENDING"],
        flags_enums=["EFLAGS"],
    )

    compared_mnemonics = [insts.CMP, insts.TEST]

    frame_pointer = regs.EBP
    stack_pointer = regs.ESP
    inst_pointer = regs.EIP


class ArchX64(ArchX86):  # type: ignore
    """Arch X64 definition"""

    address_size = 64

    frame_pointer = ArchX86.regs.RBP
    stack_pointer = ArchX86.regs.RSP
    inst_pointer = ArchX86.regs.RIP


class ArchARM(QuokkaArch):  # type: ignore
    """ArchARM definition"""

    (
        cc,
        cpsflag,
        cpsmode,
        grps,
        insts,
        mb,
        op,
        regs,
        setend,
        sft,
        sysreg,
        vd,
    ) = make_enums(
        capstone_module=capstone.arm_const,
        items=[
            "CC",
            "CPSFLAG",
            "CPSMODE",
            "GRP",
            "INS",
            "MB",
            "OP",
            "REG",
            "SETEND",
            "SFT",
            "SYSREG",
            "VECTORDATA",
        ],
        blacklist=["ENDING", "R13", "R14", "R15", "R9", "R10", "R11", "R12"],
        flags_enums=["SYSREG"],
    )

    address_size = 32
    compared_mnemonics = [
        insts.CBNZ,
        insts.CMP,
        insts.CBZ,
        insts.CMN,
    ]

    frame_pointer = regs.FP
    stack_pointer = regs.SP
    inst_pointer = regs.PC


class ArchARMThumb(ArchARM):  # type: ignore
    """Arch Arm Thumb definition"""

    thumb: bool = True


class ArchARM64(QuokkaArch):  # type: ignore
    """Arch Arm64 definition"""

    (
        at,
        barrier,
        cc,
        dc,
        ext,
        grp,
        ic,
        insts,
        op,
        prfm,
        pstate,
        regs,
        sft,
        sysreg,
        tlbi,
        vas,
        vess,
    ) = make_enums(
        capstone_module=capstone.arm64_const,
        items=[
            "AT",
            "BARRIER",
            "CC",
            "DC",
            "EXT",
            "GRP",
            "IC",
            "INS",
            "OP",
            "PRFM",
            "PSTATE",
            "REG",
            "SFT",
            "SYSREG",
            "TLBI",
            "VAS",
            "VESS",
        ],
        blacklist=["ENDING", "X16", "X17", "X29", "X30"],
        flags_enums=[],
    )

    address_size = 64
    compared_mnemonics = [
        insts.CBZ,
        insts.CBNZ,
        insts.CMP,
        insts.CCMN,
        insts.CCMP,
        insts.CMN,
        insts.TBZ,
        insts.TBNZ,
    ]

    frame_pointer = regs.FP
    stack_pointer = regs.SP  # TODO(dm)!
    inst_pointer = regs.X28


class ArchMIPS(QuokkaArch):  # type: ignore
    """Arch MIPS definition"""

    (
        grps,
        insts,
        op,
        regs,
    ) = make_enums(
        capstone_module=capstone.mips_const,
        items=[
            "GRP",
            "INS",
            "OP",
            "REG",
        ],
        blacklist=["ENDING", "INVALID"],
        flags_enums=[],
    )

    compared_mnemonics = [
        insts.CMP,
        insts.CMPI,
        insts.CMPU,
        insts.CMPGU,
        insts.CMPGDU,
    ]

    address_size = 32

    frame_pointer = regs.FP  # R30
    stack_pointer = regs.SP  # R29
    inst_pointer = regs.PC   # REG_PC


class ArchMIPS64(ArchMIPS):  # type: ignore
    """Arch MIPS64 definition"""

    address_size = 64


class ArchPPC(QuokkaArch):  # type: ignore
    """Arch PPC definition"""

    (
        bc,
        bh,
        grps,
        insts,
        op,
        regs,
    ) = make_enums(
        capstone_module=capstone.ppc_const,
        items=[
            "BC",
            "BH",
            "GRP",
            "INS",
            "OP",
            "REG",
        ],
        blacklist=["ENDING", "INVALID"],
        flags_enums=[],
    )

    compared_mnemonics = [
        insts.CMPB,
        insts.CMPD,
        insts.CMPDI,
        insts.CMPLD,
        insts.CMPLW,
        insts.CMPW,
        insts.CMPWI,
    ]

    address_size = 32

    frame_pointer = regs.R1  #
    stack_pointer = regs.R1  # GPR1
    inst_pointer = regs.CTR


class ArchPPC64(ArchPPC):  # type: ignore
    """Arch PPC64 definition"""

    address_size = 64



"""References Management

This module deals with all references between different objects in the code.

A reference is an object with the following attributes:
    - A Source
    - A Destination
    - A Type

As they are stored in a complex manner, this module deals with their resolution,
i.e. how to resolve the pointed object.

There is room for improvement here. ;)
"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
import collections
import enum

import quokka
from quokka.types import (
    AddressT,
    DefaultDict,
    Dict,
    Index,
    Iterator,
    List,
    LocationValueType,
    Mapping,
    MutableMapping,
    Optional,
    ReferenceTarget,
    ReferenceType,
    Tuple,
    Union,
)


class Reference:
    """Reference: a link between a source and a destination

    Arguments:
        source: Source of the reference
        destination: Destination (target) of the reference
        type_: Type of the reference

    Attributes:
        source: Source of the reference
        destination: Destination (target) of the reference
        type: Type of the reference
    """

    def __init__(
        self,
        source: ReferenceTarget,
        destination: ReferenceTarget,
        type_: ReferenceType,
    ) -> None:
        """Constructor"""
        self.source: ReferenceTarget = source
        self.destination: ReferenceTarget = destination
        self.type: ReferenceType = type_


class ReferencesLocation(enum.Enum):
    """Reference location

    A reference may be attached to one of the following :
        * an instruction index (that means every instance of the instruction)
        * a data index
        * a structure position (e.g. a structure or a member inside a structure)
        * an instruction : a tuple with (chunk, block, inst_index) identifying precisely
            one instruction
        * a function
        * a chunk
    """

    INSTRUCTION = "inst_idx"
    DATA = "data_idx"
    STRUCTURE = "struct_position"
    INST_INSTANCE = "instruction_position"
    FUNCTION = "function_idx"
    CHUNK = "chunk_idx"

    @staticmethod
    def from_proto(location_type: str) -> ReferencesLocation:
        """Convert reference location from proto"""
        # These are the name of the fields in the protobuf
        mapping = {
            "inst_idx": ReferencesLocation.INSTRUCTION,
            "data_idx": ReferencesLocation.DATA,
            "struct_position": ReferencesLocation.STRUCTURE,
            "instruction_position": ReferencesLocation.INST_INSTANCE,
            "function_idx": ReferencesLocation.FUNCTION,
            "chunk_idx": ReferencesLocation.CHUNK,
        }

        try:
            return mapping[location_type]
        except IndexError as exc:
            raise ValueError("Unknown location type") from exc


class References(Mapping):
    """References bucket : maintain the list of all references inside the program

    The instantiation of the class will create a mapping for every reference but not
    resolve them yet.

    This class is probably the most messy one in Quokka and needs a global refactoring.
    However, since it's somehow working, let's try to not break it yet.

    Arguments:
        program: Reference to the program

    Attributes:
        program: Reference to the program
        proto_ref: Protobuf data
        references_category: A mapping that contains every reference for each direction.
            For instance, a ref X from inst a to struct b will be stored in both the
            structure key and the instruction key

    """

    SOURCE = "source"
    DESTINATION = "destination"

    def __init__(self, program: quokka.Program) -> None:
        """Init method

        Arguments:
            program: A backref to program
        """

        # Define the update method
        def update(
            mapping: Mapping,  # type: ignore
            key: Union[Tuple[int], LocationValueType],
            value: Index,
        ) -> None:
            """Recursive method to update the `self.references_category` mapping"""
            if isinstance(key, int):
                mapping[key].append(value)
                return

            current_key: int = key[0]
            sub_key: Union[Tuple[()], Tuple[int], Tuple[int, int]] = key[1:]
            if sub_key:
                update(mapping[current_key], sub_key, value)
            elif not isinstance(mapping.get(current_key), MutableMapping):
                mapping[current_key].append(value)

        self.program: quokka.Program = program
        self.proto_ref = program.proto.references

        # Init the references mapping
        self.references_category: Dict[ReferencesLocation, DefaultDict] = {  # type: ignore
            ReferencesLocation.INSTRUCTION: collections.defaultdict(list),
            ReferencesLocation.DATA: collections.defaultdict(list),
            ReferencesLocation.STRUCTURE: collections.defaultdict(
                lambda: collections.defaultdict(list)
            ),
            ReferencesLocation.INST_INSTANCE: collections.defaultdict(
                lambda: collections.defaultdict(lambda: collections.defaultdict(list))
            ),
            ReferencesLocation.FUNCTION: collections.defaultdict(list),
            ReferencesLocation.CHUNK: collections.defaultdict(list),
        }

        reference: quokka.pb.Quokka.Reference
        for index, reference in enumerate(self.proto_ref):
            source_ref = self.references_category[self.location_type(reference.source)]
            update(source_ref, self.get_location_value(reference.source), index)

            destination_ref = self.references_category[
                self.location_type(reference.destination)
            ]
            update(
                destination_ref, self.get_location_value(reference.destination), index
            )

    @staticmethod
    def location_type(
        location: quokka.pb.Quokka.Location,
    ) -> ReferencesLocation:
        """Convert the proto location type"""
        return ReferencesLocation.from_proto(location.WhichOneof("LocationType"))

    def get_location_value(
        self, location: quokka.pb.Quokka.Location
    ) -> LocationValueType:
        """Resolve a location

        This method resolves a location for quokka.

        Arguments:
            location: Reference location

        Returns:
            A LocationTypeValue
        """
        location_type: ReferencesLocation = self.location_type(location)
        location_value: LocationValueType = getattr(location, location_type.value)

        if location_type in (
            ReferencesLocation.INSTRUCTION,
            ReferencesLocation.DATA,
            ReferencesLocation.FUNCTION,
            ReferencesLocation.CHUNK,
        ):
            return location_value

        elif location_type == ReferencesLocation.STRUCTURE:
            structure_idx: int = getattr(location_value, "structure_idx")

            if getattr(location_value, "no_member") is False:
                return structure_idx, getattr(location_value, "member_idx")

            return structure_idx, -1

        elif location_type == ReferencesLocation.INST_INSTANCE:
            return (
                getattr(location_value, "func_chunk_idx"),
                getattr(location_value, "block_idx"),
                getattr(location_value, "instruction_idx"),
            )

        raise ValueError

    def __getitem__(self, k: Index) -> Reference:
        """Lazy loading for references"""
        return self.create_reference(self.proto_ref[k])

    def __len__(self) -> int:
        """References count"""
        return len(self.proto_ref)

    def __iter__(self) -> Iterator[Dict[AddressT, Reference]]:
        """Iterator over references"""
        raise NotImplementedError

    @property
    def structures(self) -> List[quokka.Structure]:
        """Accessor to the program structures"""
        return self.program.structures

    @property
    def data(self):
        """Accessor to the program PROTO data"""
        return self.program.proto.data

    @property
    def instructions(self):
        """Accessor to the program proto instruction"""
        return self.program.proto.instructions

    def create_reference(self, reference: quokka.pb.Quokka.Reference) -> "Reference":
        """Create a reference

        Start with resolving both the source and the destination.

        Arguments:
            reference: A protobuf reference object

        Returns:
            A python reference with both source and destination resolved
        """
        source: ReferenceTarget = self.resolve_location(reference.source)
        destination: ReferenceTarget = self.resolve_location(reference.destination)

        return Reference(
            source, destination, ReferenceType.from_proto(reference.reference_type)
        )

    def find_instruction(
        self,
        instruction_identifier: "quokka.pb.Quokka.Location.InstructionIdentifier",
    ) -> Tuple[quokka.Chunk, quokka.Block, Index]:
        """Search an instruction from an instruction identifier.

        The parameters are a chunk index, a block index, and an inst index inside the
        block.

        Arguments:
            instruction_identifier: Protobuf instruction identifier

        Returns:
            A instruction tuple (Chunk, Block, Instruction Index)
        """
        chunk = self.program.get_chunk(
            instruction_identifier.func_chunk_idx, instruction_identifier.block_idx
        )
        block = chunk[chunk.index_to_address[instruction_identifier.block_idx]]
        return chunk, block, instruction_identifier.instruction_idx

    def resolve_location(self, location: quokka.pb.Quokka.Location) -> ReferenceTarget:
        """Resolve a location

        This convert a proto location to the actual instance of the object.

        Arguments:
            location: A protobuf Location (sigh) object

        Returns:
            A ReferenceTarget

        Raises:
            ValueError: When no appropriate location has been found.
        """
        location_type: ReferencesLocation = self.location_type(location)
        if location_type == ReferencesLocation.INSTRUCTION:
            raise DeprecationWarning("Not used anymore")

        elif location_type == ReferencesLocation.DATA:
            return self.program.data_holder[location.data_idx]

        elif location_type == ReferencesLocation.STRUCTURE:
            structure: quokka.Structure = self.structures[
                location.struct_position.structure_idx
            ]
            if location.struct_position.no_member is False:
                return structure[
                    structure.index_to_offset[location.struct_position.member_idx]
                ]
            return structure

        elif location_type == ReferencesLocation.INST_INSTANCE:
            return self.find_instruction(location.instruction_position)

        elif location_type == ReferencesLocation.CHUNK:
            return self.program.chunks[location.chunk_idx]

        raise ValueError("No location found")

    @staticmethod
    def get_direction(towards: bool = True) -> Tuple[str, str]:
        """Helper method to analyze if we take the reference or unwind it"""
        target = "destination" if towards else "source"
        wanted = "destination" if not towards else "source"

        return target, wanted

    def resolve_block_references(
        self,
        chunk_index: Index,
        block_index: Index,
        reference_type: ReferenceType,
        towards: bool = True,
    ) -> List[Reference]:
        """Return a list of references from/towards a block.

        Arguments:
            chunk_index: Index of the chunk
            block_index: Index of the block
            reference_type: TYpe of reference wanted (e.g. Data)
            towards: True if we want the reference *to* this block.
                False if we want the references *from* the block.

        Returns:
            A List of references matching the criteria
        """
        target, _ = self.get_direction(towards)

        return_list = []
        for reference_ids in self.references_category[ReferencesLocation.INST_INSTANCE][
            chunk_index
        ][block_index].values():
            for reference_id in reference_ids:
                reference: Reference = self[reference_id]
                if reference.type == reference_type and isinstance(
                    getattr(reference, target), tuple
                ):
                    target_loc = getattr(reference, target)
                    if (target_loc[0].proto_index, target_loc[1].proto_index) == (
                        chunk_index,
                        block_index,
                    ):
                        return_list.append(reference)

        return return_list

    def resolve_inst_instance(
        self,
        inst_tuple: Tuple[int, int, int],
        reference_type: Optional[ReferenceType] = None,
        towards: bool = True,
    ) -> List[Reference]:
        """Resolve instruction references

        Return a list of references from/towards an instruction.
        If a reference_type is specified the references will be filtered by their type

        Arguments:
            inst_tuple: A tuple (Chunk Index, Block Index, Instruction Index)
            reference_type: Reference type to consider (filter result)
            towards: In which sense to search the reference?

        Returns:
            A list of reference matching the criteria
        """
        return_list = []
        target, _ = self.get_direction(towards)

        for reference_idx in self.references_category[ReferencesLocation.INST_INSTANCE][
            inst_tuple[0]
        ][inst_tuple[1]][inst_tuple[2]]:
            reference: "Reference" = self[reference_idx]
            if (not reference_type or reference.type == reference_type) and isinstance(
                getattr(reference, target), tuple
            ):
                target_loc = getattr(reference, target)
                if (
                    target_loc[0].proto_index,
                    target_loc[1].proto_index,
                    target_loc[2],
                ) == inst_tuple:
                    return_list.append(reference)

        return return_list

    def resolve_calls(
        self, chunk: quokka.Chunk, towards: bool = True
    ) -> List[Union[Tuple[quokka.Chunk, quokka.Block, int], quokka.Chunk]]:
        """Resolve calls to a chunk

        This method resolves calls initiated or received from every instruction in a
        chunk. Of note, a call must be from an Instruction Tuple to another.

        Arguments:
            chunk: Target chunk
            towards: Do we look at calls towards this chunk {X,Y,Z} -> Chunk or at calls
                from this chunk Chunk -> {X,Y,Z} ?

        Returns:
            A list of objects that are either tuple (Chunk, Block, Instruction Index) or
            Chunk (only for fake chunks that are created from an imported function)
        """
        return_list = []
        target, wanted = self.get_direction(towards)

        for block_index in self.references_category[ReferencesLocation.INST_INSTANCE][
            chunk.proto_index
        ]:
            # For super-chunks, filter out block that do not belong to this chunk
            # anymore
            if block_index not in chunk.index_to_address:
                continue

            block_references = self.references_category[
                ReferencesLocation.INST_INSTANCE
            ][chunk.proto_index][block_index]
            for references in block_references.values():
                for reference_idx in references:
                    reference: "Reference" = self[reference_idx]
                    if reference.type == ReferenceType.CALL and isinstance(
                        getattr(reference, target), tuple
                    ):
                        target_loc = getattr(reference, target)
                        if target_loc[0].proto_index == chunk.proto_index:
                            location = getattr(reference, wanted)
                            # Fix: Only returns valid tuple
                            # TODO(dm): check why sometimes a Data might be returned
                            if isinstance(location, (tuple, quokka.Chunk)):
                                return_list.append(location)

        return return_list

    def resolve_data(
        self,
        data_index: Index,
        reference_type: Union[ReferenceType, None] = None,
    ) -> List[Reference]:
        """Resolve data references

        Returns a list of reference towards a data
        If a reference_type is specified the references will be filtered by their type

        Arguments:
            data_index: Index of the data in the protobuf
            reference_type: Type of reference

        Returns:
            A list of reference matching the criteria
        """

        references: List[Reference] = []
        for reference_idx in self.references_category[ReferencesLocation.DATA][
            data_index
        ]:
            reference = self[reference_idx]

            if reference_type is not None and reference_type != reference.type:
                continue

            references.append(reference)

        return references



"""Data management.

A data is a piece of information that isn't code.
"""

#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
import logging

import quokka

from quokka.types import (
    AddressT,
    Any,
    DataType,
    Index,
    List,
    Mapping,
    Optional,
)

logger = logging.getLogger(__name__)


class Data:
    """Base class for data.

    All data have at least a type and a value.
    They are referenced inside the program by and to other data and code.

    Parameters:
        proto_index: Index in the protobuf
        data: Protobuf value of the data.
        program: Program backref

    Attributes:
        proto_index: Index in the protobuf
        address: Data address
        type: Data type
        program: Reference to the Program
        is_initialized: Is the data initialized?
        size: Data size (depends on the type usually)
        name: Data name (if any)
    """

    def __init__(
        self, proto_index: Index, data: "quokka.pb.Quokka.Data", program: quokka.Program
    ):
        """Constructor"""
        self.proto_index: Index = proto_index
        self.address: AddressT = program.addresser.absolute(data.offset)
        self.type: "DataType" = DataType.from_proto(data.type)
        self.program: quokka.Program = program

        self.is_initialized: bool = not data.not_initialized

        self.size: Optional[int] = (
            data.size if data.WhichOneof("DataSize") != "no_size" else None
        )
        self._value: Optional[str] = (
            self.program.proto.string_table[data.value_index]
            if data.value_index > 0
            else None
        )
        self.name: Optional[str] = (
            self.program.proto.string_table[data.name_index]
            if data.name_index > 0
            else None
        )

    def __eq__(self, other: Any) -> bool:
        """Check equality between two Data instances"""
        return type(other) is type(self) and other.proto_index == self.proto_index

    @property
    def value(self) -> Any:
        """Data value.

        The value is read in the program binary file.
        """

        # Uninitialized memory
        if not self.is_initialized:
            return None

        address = self.program.addresser.file(self.address)

        if self.type in (
            DataType.ALIGN,
            DataType.POINTER,
            DataType.STRUCT,
            DataType.UNKNOWN,
        ):
            return self._value

        if self.type == DataType.ASCII:
            try:
                return self.program.executable.read_data(
                    address, self.type, size=self.size
                )
            except quokka.exc.NotInFileError:
                logger.error("Try to read a string which is not in file")
                return ""
        else:
            return self.program.executable.read_data(address, self.type)

    @property
    def references(self) -> List[quokka.Reference]:
        """References to/from this data"""
        return self.program.references.resolve_data(self.proto_index)

    @property
    def code_references(self) -> List[quokka.Reference]:
        """Returns code referencing this Data"""
        return [ref for ref in self.references if isinstance(ref.destination, tuple)]

    @property
    def data_references(self) -> List[quokka.Reference]:
        """Returns data references to/from this Data"""
        return [ref for ref in self.references if isinstance(ref.destination, Data)]


class DataHolder(Mapping):
    """Data bucket

    All the data of the program are referenced in this bucket and allow to store them
    only once.

    Attributes:
        proto_data: The protobuf data themselves
        program: A reference to the Program

    Arguments:
        proto: The protobuf data
        program: The program

    TODO:
        Type hinting for proto parameter (RepeatedCompositeFieldContainer)
    """

    def __init__(self, proto, program: quokka.Program):
        """Init method

        Arguments:
            proto: List of data in the protobuf
            program: Backref to the program
        """
        self.proto_data = proto.data
        self.program: quokka.Program = program

    def __setitem__(self, key: Index, value: Data) -> None:
        """Set a data"""
        raise ValueError("Should not be accessed")

    def __delitem__(self, value: Index) -> None:
        """Remove a data from the bucket"""
        raise ValueError("Should not be accessed")

    def __getitem__(self, key: Index) -> Data:
        """Get a data from the bucket.

        Arguments:
            key: Data Index

        Returns:
            A Data
        """
        return Data(key, self.proto_data[key], self.program)

    def get_data(self, address: AddressT) -> Data:
        """Find a data by address

        Iterates over the data to find the one at a specified offset

        Arguments:
            address: Offset to query

        Returns:
            A Data

        Raises:
            ValueError: if no data is found
        """

        # We have to iterate over every data because they are not sorted by offset
        for index, data_proto in enumerate(self.proto_data):
            if data_proto.offset + self.program.base_address == address:
                return self[index]

        raise ValueError(f"No data at offset 0x{address:x}")

    def __len__(self) -> int:
        """Number of data in the program"""
        return len(self.proto_data)

    def __iter__(self):
        """Do not allow the iteration over the data"""
        raise ValueError("Should not be accessed")



"""Calling conventions"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from quokka.analysis import ArchX86, ArchX64, ArchARM, ArchARM64
from quokka.types import List, RegType


# Inspired from https://github.com/angr/angr/blob/master/angr/calling_conventions.py


class CallingConvention:
    """Base class for a calling convention"""

    name: str
    argument_registers: List[RegType]
    floating_point_registers: List[RegType]
    caller_saved_registers: List[RegType]
    callee_saved: bool = False

    return_value: RegType
    overflow_return: RegType


class CCdecl(CallingConvention):
    """Cdecl calling convention"""

    name = "cdecl"
    argument_registers: List[RegType] = []  # All arguments are on the stack
    floating_point_registers: List[RegType] = []  # Same
    caller_saved_registers = [
        ArchX86.regs.EAX,
        ArchX86.regs.ECX,
        ArchX86.regs.EDX,
    ]

    return_value = ArchX86.regs.EAX
    overflow_return = ArchX86.regs.EDX


class Stdcall(CCdecl):
    """Stdcall calling convention"""

    name = "stdcall"
    callee_saved = True


class Fastcall(CallingConvention):
    """Fastcall calling convention"""

    name = "fastcall"
    argument_registers = [ArchX86.regs.ECX, ArchX86.regs.EDX]
    floating_point_registers = [ArchX86.regs.ST0, ArchX86.regs.ST1]  # TODO(dm) check
    caller_saved_registers = [
        ArchX86.regs.EBX,
        ArchX86.regs.EBX,
        ArchX86.regs.ESI,
        ArchX86.regs.EDI,
    ]

    return_value = ArchX86.regs.EAX
    overflow_return = ArchX86.regs.EDX


class MicrosoftAMD64(CallingConvention):
    """Microsoft 64 calling convention"""

    name = "ms"
    argument_registers = [
        ArchX64.regs.RCX,
        ArchX64.regs.RDX,
        ArchX64.regs.R8,
        ArchX64.regs.R9,
    ]

    floating_point_registers = [
        ArchX64.regs.XMM0,
        ArchX64.regs.XMM1,
        ArchX64.regs.XMM2,
        ArchX64.regs.XMM3,
    ]

    return_value = ArchX64.regs.RAX
    overflow_return = ArchX64.regs.RDX


class SystemVAMD(CallingConvention):
    """SysV calling convention"""

    name = "sysv"
    argument_registers = [
        ArchX64.regs.RDI,
        ArchX64.regs.RSI,
        ArchX64.regs.RDX,
        ArchX64.regs.RCX,
        ArchX64.regs.R8,
        ArchX64.regs.R9,
    ]

    floating_point_registers = [
        ArchX64.regs.XMM0,
        ArchX64.regs.XMM1,
        ArchX64.regs.XMM2,
        ArchX64.regs.XMM3,
        ArchX64.regs.XMM4,
        ArchX64.regs.XMM5,
        ArchX64.regs.XMM6,
        ArchX64.regs.XMM7,
    ]

    caller_saved_registers = argument_registers + [
        ArchX64.regs.R10,
        ArchX64.regs.R11,
        ArchX64.regs.RAX,
    ]

    return_value = ArchX64.regs.RAX
    overflow_return = ArchX64.regs.RDX


class ARMCC(CallingConvention):
    """AAPCS calling convention for ARM"""

    name = "aapcs"
    argument_registers = [
        ArchARM.regs.R0,
        ArchARM.regs.R1,
        ArchARM.regs.R2,
        ArchARM.regs.R3,
    ]
    floating_point_registers: List[RegType] = []  # TODO
    caller_saved_registers: List[RegType] = []
    return_value = ArchARM.regs.R0


class ARM64CC(CallingConvention):
    """AAPCS calling convention for Aarch64"""

    name = "aapcs"
    argument_registers = [
        ArchARM64.regs.X0,
        ArchARM64.regs.X1,
        ArchARM64.regs.X2,
        ArchARM64.regs.X3,
        ArchARM64.regs.X4,
        ArchARM64.regs.X5,
        ArchARM64.regs.X6,
        ArchARM64.regs.X7,
    ]

    floating_point_registers: List[RegType] = []  # TODO
    caller_saved_registers: List[RegType] = []
    return_value = ArchARM64.regs.X0



"""Segment management"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import quokka

from quokka.types import (
    AddressT,
    SegmentType,
)


class Segment:
    """Segment

    This class represents IDA segments.

    Arguments:
        segment: Segment protobuf information
        program: Program back reference

    Attributes:
        name: Segment name
        start: Segment starting address
        permissions: Segment permissions
        size: Segment size
        type: Segment type
        program: Program reference
        file_offset: Segment offset in the file (if appropriate)

    """

    def __init__(
        self,
        segment: "quokka.pb.Quokka.Segment",
        program: quokka.Program,
    ):
        """Constructor"""
        self.name: str = segment.name
        self.start: AddressT = segment.start_addr
        self.permissions: int = segment.permissions
        self.size: int = segment.size
        self.type: "SegmentType" = SegmentType.from_proto(segment.type)

        self.program: quokka.Program = program

        self.file_offset: int = -1
        if segment.no_offset is False:
            self.file_offset = segment.file_offset - self.start

    @property
    def end(self) -> AddressT:
        """End address of the segment"""
        return self.start + self.size

    def writable(self) -> bool:
        """Is the segment writable?"""
        return self.permissions & 0x2 > 0

    def executable(self) -> bool:
        """Is the segment executable?"""
        return self.permissions & 0x1 > 0

    def readable(self) -> bool:
        """Is the segment readable?"""
        return self.permissions & 0x4 > 0

    def in_segment(self, addr: int) -> bool:
        """Does `addr` belong to this segment ?"""
        return self.start <= addr < self.start + self.size



"""PyPCode integration"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import itertools
import logging

import pypcode

import quokka
import quokka.analysis
from quokka.types import Any, Dict, Endianness, List, Sequence, Type, Optional

logger: logging.Logger = logging.getLogger(__name__)


def get_arch_from_string(target_id: str) -> pypcode.ArchLanguage:
    """Find the architecture for an arch based on the target identification

    Arguments:
        target_id: Identifier of the architecture

    Raises:
        PypcodeError: if the architecture is not found

    Returns:
        The appropriate ArchLang
    """
    pcode_arch: pypcode.Arch
    for pcode_arch in pypcode.Arch.enumerate():
        for lang in pcode_arch.languages:
            if lang.id == target_id:
                return lang

    raise quokka.PypcodeError("Unable to find the appropriate arch: missing lang")


def get_pypcode_context(
        arch: Type[quokka.analysis.QuokkaArch],
        endian: Type[Endianness] = Endianness.LITTLE_ENDIAN
) -> pypcode.Context:
    """Convert an arch from Quokka to Pypcode

    For the moment, only the arch described in quokka.analysis are supported.
    This method is a bit slow because enum are generated by pypcode on the fly but should
    be executed only once.

    Arguments:
        arch: Quokka program architecture
        endian: Architecture endianness

    Raises:
        PypcodeError: if the conversion for arch is not found

    Returns:
        A pypcode.Context instance
    """
    names: Dict[Type[quokka.analysis.arch.QuokkaArch], str] = {
        quokka.analysis.ArchX64: "x86:LE:64:default",
        quokka.analysis.ArchX86: "x86:LE:32:default",
        quokka.analysis.ArchARM: "ARM:LE:32:v8",
        quokka.analysis.ArchARM64: "AARCH64:LE:64:v8A",
        quokka.analysis.ArchARMThumb: "ARM:LE:32:v8T",
        quokka.analysis.ArchMIPS: "MIPS:LE:32:default",
        quokka.analysis.ArchMIPS64: "MIPS:LE:64:default",
        quokka.analysis.ArchPPC: "PowerPC:LE:32:default",
        quokka.analysis.ArchPPC64: "PowerPC:LE:64:default",
    }

    try:
        target_id = names[arch]
    except KeyError as exc:
        raise quokka.PypcodeError(
            "Unable to find the appropriate arch: missing id"
        ) from exc

    if endian == Endianness.BIG_ENDIAN:
        target_id = target_id.replace(":LE:", ":BE:")

    pcode_arch = get_arch_from_string(target_id)
    return pypcode.Context(pcode_arch)


def update_pypcode_context(program: quokka.Program, is_thumb: bool) -> pypcode.Context:
    """Return an appropriate pypcode context for the decoding

    For ARM architecture, if the block starts with a Thumb instruction, we must use
    a different pypcode Context.

    We use the boolean `is_thumb` directly to allow caching of the call here because it
    is costly to generate the context.

    Arguments:
        program: Program to consider
        is_thumb: Is the instruction a thumb one?

    Returns:
        The correct pypcode context
    """

    if (
        program.arch
        in (
            quokka.analysis.ArchARM,
            quokka.analysis.ArchARM64,
            quokka.analysis.ArchARMThumb,
        )
        and is_thumb
    ):
        return get_pypcode_context(quokka.analysis.ArchARMThumb)

    return program.pypcode


def pypcode_decode_block(block: quokka.Block) -> List[pypcode.PcodeOp]:
    """Decode a block at once.

    This method decode a block of instructions using Pypcode context all at once.
    This is faster than multiple calls to the decode at the instruction level.

    Arguments:
        block: Block to decode

    Returns:
        A list of pcode operations
    """

    # Fast guard, empty blocks do not have any Pcode operations
    first_instruction: Optional[quokka.Instruction] = next(block.instructions, None)
    if first_instruction is None:
        return []

    # Retrieve the context from the instruction
    context: pypcode.Context = update_pypcode_context(
        block.program, first_instruction.thumb
    )

    try:
        # Translate
        translation = context.translate(
            block.bytes,  # buf
            block.start,  # base_address
            0,  # max_bytes
            0,  # max_instructions
        )
        return translation.ops

    except pypcode.BadDataError as e:
        logger.error(e)
        raise quokka.PypcodeError(f"Decoding error for block at 0x{block.start:x} (BadDataError)")
    except pypcode.UnimplError as e:
        logger.error(e)
        raise quokka.PypcodeError(f"Decoding error for block at 0x{block.start:x} (UnimplError)")


def pypcode_decode_instruction(
    inst: quokka.Instruction,
) -> Sequence[pypcode.PcodeOp]:
    """Decode an instruction using Pypcode

    This will return the list of Pcode operations done for the instruction.
    Note that a (binary) instruction is expected to have several pcode instructions
    associated. When decoding a single instruction IMARK instructions are excluded!

    Arguments:
        inst: Instruction to translate

    Raises:
        PypcodeError: if the decoding fails

    Returns:
        A sequence of PcodeOp
    """

    context: pypcode.Context = update_pypcode_context(inst.program, inst.thumb)
    try:
        translation = context.translate(
            inst.bytes,  # buf
            inst.address,  # base_address
            0,  # max_bytes
            1,  # max_instructions
        )

        return [x for x in translation.ops if x.opcode != pypcode.OpCode.IMARK]

    except pypcode.BadDataError as e:
        logger.error(e)
        raise quokka.PypcodeError(f"Unable to decode instruction (BadDataError)")
    except pypcode.UnimplError as e:
        logger.error(e)
        raise quokka.PypcodeError(f"Unable to decode instruction (UnimplError)")



"""Executable: management of the binary file in itself."""

#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pathlib
import struct

from quokka.types import DataType, Endianness, Literal, Optional, Union


class Executable:
    """The executable class is used to interact with the binary file.

    It handles access to the binary file itself (not the exported) for reads.

    Note: The binary is read only once and stored in memory. This is done for
    performance purposes but does not cope well with low RAM systems and/or huge
    binaries.

    Arguments:
        path: Path towards the executable file
        endianness: How are stored the data

    Attributes:
        exec_file: Path towards the executable file
        endianness: Binary endianness
        content: Bytes of the binary

    Raises:
        ValueError: If the file is not found
    """

    def __init__(self, path: Union[str, pathlib.Path], endianness: Endianness):
        """Constructor"""
        try:
            with open(path, "rb") as file:
                self.content: bytes = file.read()

        except FileNotFoundError:
            raise ValueError("File not found")

        self.exec_file: pathlib.Path = pathlib.Path(path)
        self.endianness: Endianness = endianness

    def read(self, offset: int, size: int) -> bytes:
        """Read `size` at `offset` in the file.

        This method should not be used directly and considered as part of a private API.
        The preferred method are read_bytes / read_string .

        Arguments:
            offset: File offset
            size: Read size

        Returns:
            The content that has been read

        Raises:
            ValueError: when the value is not in the file
        """
        try:
            return self.content[offset : offset + size]
        except IndexError as exc:
            raise ValueError(f"Content not found at offset {offset}") from exc

    def read_string(self, offset: int, size: Optional[int] = None) -> str:
        """Read a string in the file.

        If the size is not given, Quokka will try to read the string until the
        first null byte. That works only for null-terminated strings.

        If the string is null terminated, remove the trailing 0.

        Arguments:
            offset: String file offset
            size: String size if known.

        Returns:
            The decoded string

        Raises:
          ValueError: If the string is not found nor decoded.
        """

        if size is not None:
            try:
                string = self.read(offset, size).decode("utf-8")
            except UnicodeDecodeError as exc:
                raise ValueError("Unable to read or decode the string.") from exc

        else:
            try:
                null_byte = self.content.index(b"\x00", offset)
            except ValueError as exc:
                raise ValueError(
                    "String is not null-terminated and size was not given"
                ) from exc

            string = self.content[offset:null_byte].decode("utf-8")

        # FIX: When returning a single character string, it does not end with a '\0'
        if len(string) > 1 and string.endswith("\x00"):
            return string[:-1]

        return string

    def read_data(
        self, offset: int, data_type: DataType, size: Optional[int] = None
    ) -> Union[int, float, str]:
        """Read the data value.

        If the size is not specified, it is inferred from the data type.

        Arguments:
            offset: Data file offset
            data_type: Data type
            size: Read size

        Returns:
            The data value
        """

        # Read an int of size `read_size`
        def read_int(read_size: int) -> int:
            """Read an integer from the binary"""
            return int.from_bytes(self.read_bytes(offset, read_size), endianness)

        endianness: Literal["big", "little"]
        if self.endianness == Endianness.BIG_ENDIAN:
            endianness = "big"
            endianness_sign = ">"
        else:
            endianness = "little"
            endianness_sign = "<"

        if data_type == DataType.ASCII:
            if size is None:
                raise ValueError("No size specified when reading a DataType.ASCII")
            return self.read_string(offset, size)
        elif data_type == DataType.BYTE:
            return read_int(1 if size is None else size)
        elif data_type == DataType.WORD:
            return read_int(2 if size is None else size)
        elif data_type == DataType.DOUBLE_WORD:
            return read_int(4 if size is None else size)
        elif data_type == DataType.QUAD_WORD:
            return read_int(8 if size is None else size)
        elif data_type == DataType.OCTO_WORD:
            return read_int(16 if size is None else size)
        elif data_type == DataType.FLOAT:
            s = 4 if size is None else size
            return struct.unpack(f"{endianness_sign}f", self.read_bytes(offset, s))
        elif data_type == DataType.DOUBLE:
            s = 8 if size is None else size
            return struct.unpack(f"{endianness_sign}d", self.read_bytes(offset, s))
        else:
            raise NotImplementedError(
                f"Cannot read {data_type}. DataType not implemented."
            )

    def read_bytes(self, offset: int, size: int) -> bytes:
        """Read one (or more) byte(s) in the file at `offset`.

        This is mostly used to read instructions.

        Arguments:
            offset: File offset to read
            size: Number of bytes to read

        Returns:
            The bytes values
        """
        return self.read(offset, size)



"""Utilities functions for the analysis"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
import networkx as nx

import quokka
from quokka.types import Union


def split_chunk(
    chunk: quokka.Chunk,
) -> Union[quokka.Chunk, quokka.SuperChunk]:
    """Split a chunk if it is composed of multiple components.

    If a chunk is composed of multiple connected components, we want to split them so
    some analysis may be performed.

    Arguments:
        chunk: A chunk to split

    Returns:
        Either a Chunk or a SuperChunk
    """
    components = list(nx.connected_components(nx.Graph(chunk.graph)))
    if len(components) <= 1:
        return chunk

    return quokka.function.SuperChunk(chunk, components)



"""Structure management"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
import weakref
import quokka

from quokka.types import (
    DataType,
    Dict,
    List,
    Optional,
    StructureType,
)


class StructureMember:
    """StructureMember

    This class represents structure members (fields).

    Arguments:
        member: Protobuf data
        structure: Reference to the parent structure

    Attributes:
        name: Member name
        size: Member size (if known)
        type: Member data type
        value: Member value
        comments: Member comments
    """

    def __init__(
        self,
        member: "quokka.pb.Quokka.Structure.Member",
        structure: Structure,
    ) -> None:
        """Constructor"""
        self.name: str = member.name
        self.type: DataType = DataType.from_proto(member.type)
        self.size: int = member.size
        self.value: Optional[int] = member.value if member.value != 0 else None
        self._structure: weakref.ref[Structure] = weakref.ref(structure)

        self.comments: List[str] = []

    @property
    def structure(self) -> Structure:
        """Back reference to the parent structure"""
        return self._structure()


class Structure(dict):
    """Structure

    All IDA structure are merged inside this class (Enum, Structure, Union).

    Arguments:
        structure: Structure protobuf data
        program: Program back reference

    Attributes:
        program: Program backreference
        name: Structure name
        size: Structure size (if known)
        type: Structure type
        index_to_offset: Mapping from offsets to structure members
        comments: Structure comments
    """

    def __init__(
        self,
        structure: "quokka.pb.Quokka.Structure",
        program: quokka.Program,
    ) -> None:
        """Constructor"""
        super(dict, self).__init__()
        self.program: quokka.Program = program
        self.name: str = structure.name
        self.size: Optional[int] = (
            structure.size if structure.variable_size is False else 0
        )
        self.type = StructureType.from_proto(structure.type)

        self.index_to_offset: Dict[int, int] = {}
        for index, member in enumerate(structure.members):
            self[member.offset] = StructureMember(member, self)
            self.index_to_offset[index] = member.offset

        self.comments: List[str] = []



"""Program

This is the main class of Quokka.
It deals with the most common abstraction, the Program.
"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import collections
from functools import cached_property
from itertools import product
import logging
import os
import pathlib
import subprocess
from typing import TYPE_CHECKING

import capstone
import networkx

import quokka
import quokka.analysis
import quokka.backends

from quokka.types import (
    AddressT,
    Dict,
    Endianness,
    ExporterMode,
    FunctionType,
    Index,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    Union,
)

if TYPE_CHECKING:
    import pypcode


class Program(dict):
    """Program

    The program is `quokka` main abstraction.
    It represents the full binary and is in itself a mapping of functions.

    Arguments:
        export_file: Path towards the export file (e.g. .quokka)
        exec_path: Path towards the binary file

    Attributes:
        proto: Contains the protobuf data. This should not be used directly. However, if
            you don't find another way of accessing some information, feel  free to
            open an issue.
        export_file: The path to the export file (e.g. the .quokka)
        mode: Export mode (LIGHT, NORMAL or FULL)
        base_address: Program base address
        addresser: Utility to convert the program offsets into file-offsets
        isa: Instruction set
        address_size: Default pointer size
        arch: Program architecture
        endianness: Program endianness
        chunks: A mapping of chunks
        executable: An object to manage the binary file
        references: The reference manager
        data_holder: The data manager
        fun_names: A mapping of function names to functions

    Raises:
        QuokkaError: If the loading is not successful.

    """

    logger: logging.Logger = logging.getLogger(__name__)

    def __init__(
        self, export_file: Union[pathlib.Path, str], exec_path: Union[pathlib.Path, str]
    ):
        """Constructor"""
        super(dict, self).__init__()

        self.proto: quokka.pb.Quokka = quokka.pb.Quokka()
        self.export_file: pathlib.Path = pathlib.Path(export_file)
        with open(self.export_file, "rb") as fd:
            self.proto.ParseFromString(fd.read())

        # Export mode
        self.mode: ExporterMode = ExporterMode.from_proto(self.proto.exporter_meta.mode)

        # Version checking
        # A change in the major version might break backward compatibility
        proto_version = quokka.parse_version(self.proto.exporter_meta.version)
        current_version = quokka.parse_version(quokka.__version__)
        if proto_version[0] != current_version[0]:
            self.logger.warning(
                "The exported file has been generated by a different version of Quokka."
                f" The file has been generated by Quokka {self.proto.exporter_meta.version}"
                f" while you are using {quokka.__version__}"
            )
        elif self.proto.exporter_meta.version != quokka.__version__:
            self.logger.debug(
                "Version mismatch detected but still compatible with the exported file."
                f" The file has been generated by Quokka {self.proto.exporter_meta.version}"
                f" while you are using {quokka.__version__}"
            )

        # Check if the hashes matches between the export file and the exec
        if not quokka.check_hash(self.proto.meta.hash, pathlib.Path(exec_path)):
            self.logger.error("Hash does not match with file.")
            raise quokka.QuokkaError("Hash mismatch")

        self.base_address: AddressT = self.proto.meta.base_addr
        self.addresser = quokka.Addresser(self, self.base_address)

        self.isa: quokka.analysis.ArchEnum = quokka.get_isa(self.proto.meta.isa)
        self.address_size: int = quokka.convert_address_size(
            self.proto.meta.address_size
        )
        self.arch: Type[quokka.analysis.QuokkaArch] = quokka.get_arch(
            self.isa, self.address_size
        )

        self.endianness: Endianness = Endianness.from_proto(self.proto.meta.endianess)

        self.chunks: Dict[int, Union[quokka.Chunk, quokka.SuperChunk]] = {}

        self.executable = quokka.Executable(exec_path, self.endianness)
        self.references = quokka.References(self)
        self.data_holder = quokka.DataHolder(self.proto, self)

        # Chunks
        for chunk_index, _ in enumerate(self.proto.function_chunks):
            chunk = quokka.Chunk(chunk_index, program=self)

            if chunk.fake:
                chunk = quokka.analysis.split_chunk(chunk)

            self.chunks[chunk_index] = chunk

        # Functions
        self.fun_names: Dict[str, quokka.Function] = {}
        for func in self.proto.functions:
            function = quokka.Function(func, self)
            self[function.start] = function
            if function.name not in self.fun_names:
                self.fun_names[function.name] = function
            else:
                if function.type == self.fun_names[function.name]:
                    self.logger.warning("Found two functions with the same name.")
                else:
                    if function.type == FunctionType.NORMAL:
                        self.fun_names[function.name] = function

    def __hash__(self) -> int:
        """Hash of the Program (use the hash from the exported file)"""
        return int(self.proto.meta.hash.hash_value, 16)

    @property
    def name(self) -> str:
        """Returns the underlying binary name"""
        return self.proto.meta.executable_name

    @property
    def hash(self) -> str:
        """Returns the hash value of the binary (either sha256 or MD5)."""
        return self.proto.meta.hash.hash_value

    @cached_property
    def capstone(self) -> capstone.Cs:
        """Compute a capstone context"""
        return quokka.backends.get_capstone_context(self.arch, self.endianness)

    @cached_property
    def call_graph(self) -> networkx.DiGraph:
        """Compute the Call Graph of the binary

        Every node in the call graph is a chunk (and not a function).

        :return: A Call Graph (a networkx DiGraph)
        """
        call_graph: "networkx.DiGraph" = networkx.DiGraph()

        chunk: Union[quokka.Chunk, quokka.SuperChunk]
        for chunk in self.chunks.values():
            if isinstance(chunk, quokka.Chunk):
                call_graph.add_node(chunk.start)
                call_graph.add_edges_from(product((chunk.start,), chunk.calls))
            else:  # Super Chunks
                for small_chunk in chunk.starts.values():
                    call_graph.add_edges_from(
                        product((small_chunk,), small_chunk.calls)
                    )

        return call_graph

    @cached_property
    def pypcode(self) -> pypcode.Context:
        """Generate the Pypcode context."""
        from quokka.backends.pypcode import get_pypcode_context

        return get_pypcode_context(self.arch, self.endianness)

    @cached_property
    def structures(self) -> List[quokka.Structure]:
        """Structures accessor

        Allows to retrieve the different structures of a program (as defined by the
        disassembler).

        Returns:
            A list of structures
        """
        structures = [
            quokka.Structure(structure, self) for structure in self.proto.structs
        ]
        return structures

    @property
    def strings(self) -> Iterable[str]:
        """Program strings

        Retrieves all the strings used in the program.

        Returns:
            A list of strings.
        """
        # Do not use the empty string (the first one)
        return self.proto.string_table[1:]

    @cached_property
    def segments(self) -> List[quokka.Segment]:
        """Returns the list of segments defined in the program."""
        return [quokka.Segment(segment, self) for segment in self.proto.segments]

    def get_instruction(self, address: AddressT) -> quokka.Instruction:
        """Get an instruction by its address

        Note: the address must be the head of the instruction.

        TODO(dm): Improve the algorithm because the chunks are sorted (use bisect)

        Arguments:
            address: AddressT: Address to query

        Returns:
            A `quokka.Instruction`

        Raises:
            IndexError: When no instruction is found at this address
        """
        for chunk in self.chunks.values():
            if chunk.in_chunk(address):
                try:
                    return chunk.get_instruction(address)
                except IndexError:
                    pass

        raise IndexError(f"No instruction at address 0x{address:x}")

    def get_function(
        self, name: str, approximative: bool = True, normal: bool = False
    ) -> quokka.Function:
        """Find a function in a program by its name

        Arguments:
            name: Function name
            approximative: Should the name exactly match or allow partial matches?
            normal: Return only FunctionType.NORMAL functions

        Returns:
            A function matching the research criteria

        Raises:
            ValueError: When no function is found
        """
        if approximative is False:
            try:
                return self.fun_names[name]
            except KeyError as exc:
                raise ValueError("Missing function") from exc

        for function_name, function in self.fun_names.items():
            # TODO(dm) Improve this
            if name in function.name and (
                not normal or function.type == FunctionType.NORMAL
            ):
                return self.fun_names[function_name]

        raise ValueError("Unable to find an appropriate function")

    def get_segment(self, address: AddressT) -> quokka.Segment:
        """Get a `Segment` by an address

        The address must be in [segment.start, segment.end) to be found.

        Arguments:
            address: Segment's address

        Returns:
            The corresponding Segment

        Raises:
            KeyError: When the segment is not found
        """
        for segment in self.segments:
            if segment.in_segment(address):
                return segment

        raise KeyError(f"No segment has been found for address {address:#x}")

    @cached_property
    def func_chunk_index(self) -> Dict[Index, List[quokka.Function]]:
        """Returns the list of functions attached to a chunk.

        This method allows to find all the functions using a specific chunk.
        However, it is mostly an internal method and should not be directly used by a
        user. Instead, use `get_function_by_chunk`.

        Returns:
            A mapping of ChunkIndex to a list of Function.
        """
        func_chunk_index = collections.defaultdict(list)
        for function in self.values():
            for chunk_proto_index in function.index_to_address:
                func_chunk_index[chunk_proto_index].append(function)

        return func_chunk_index

    def get_function_by_chunk(self, chunk: quokka.Chunk) -> List[quokka.Function]:
        """Retrieves all the functions where `chunk` belongs.

        Arguments:
            chunk: Chunk to search for

        Returns:
            A list of corresponding functions

        Raises:
            IndexError: When no function is found for the chunk.

        """
        functions = self.func_chunk_index[chunk.proto_index]
        if not functions:
            raise IndexError(
                "No function has been found for the chunk. "
                "This is probably a Quokka bug and should be reported."
            )

        return functions

    def get_first_function_by_chunk(
        self, chunk: quokka.Chunk
    ) -> Optional[quokka.Function]:
        """Return the first function found when searching for a chunk.

        Arguments:
            chunk: Chunk belonging to the function

        Returns:
          A function in which `chunk` belongs

        Raises:
            FunctionMissingError: No function has been found for the chunk
        """
        try:
            return self.get_function_by_chunk(chunk)[0]
        except IndexError:
            raise quokka.FunctionMissingError("Missing function from chunk")

    def get_chunk(
        self, chunk_index: Index, block_index: Optional[Index] = None
    ) -> quokka.Chunk:
        """Get a `Chunk`

        If the candidate Chunk is a SuperChunk, this method will resolve it to find the
        appropriate chunk (given a block index).

        Arguments:
            chunk_index: Chunk index
            block_index: Used to resolve SuperChunks

        Returns:
            A Chunk matching the criteria

        Raises:
            ChunkMissingError: When no chunk has been found
        """
        chunk = self.chunks.get(chunk_index, None)
        if isinstance(chunk, quokka.Chunk):
            return chunk

        if isinstance(chunk, quokka.SuperChunk):
            if block_index is None:
                raise quokka.ChunkMissingError(
                    "Unable to find the chunk requested because its a super chunk"
                )

            return chunk.get_chunk_by_index(chunk_index, block_index)

        raise quokka.ChunkMissingError("Unable to find the chunk, index unknown")

    def iter_chunk(
        self, chunk_types: Optional[List[FunctionType]] = None
    ) -> Iterator[quokka.Chunk]:
        """Iterate over all the chunks in the program.

        If a `SuperChunk` is found, it will split it and return individual chunks.
        By default, it iterates over all the chunks, even extern functions.

        Arguments:
            chunk_types: Allow list of chunk types. By default, it retrieves every
                chunk.

        Yields:
            All the chunks in the program.
        """

        if chunk_types is None:
            chunk_types = list(FunctionType)

        chunk: quokka.Chunk
        for chunk in self.chunks.values():
            if isinstance(chunk, quokka.SuperChunk):
                inner_chunk: quokka.Chunk
                for inner_chunk in chunk.values():
                    if inner_chunk.chunk_type in chunk_types:
                        yield inner_chunk
            else:
                if chunk.chunk_type in chunk_types:
                    yield chunk

    def read_bytes(self, v_addr: AddressT, size: int) -> bytes:
        """Read raw bytes from a virtual address

        Arguments:
            v_addr: Virtual address of the data to read
            size: Size of the data to read

        Returns:
            The raw data at the specified address
        """

        if (offset := v_addr - self.base_address) < 0:
            raise ValueError("Address outside virtual address space.")
        return self.executable.read_bytes(offset, size)

    def get_data(self, address: AddressT) -> quokka.Data:
        """Get data by address

        Arguments:
            address: Address to query

        Returns:
            A data at the address
        """
        return self.data_holder.get_data(address)

    def __repr__(self) -> str:
        """Program representation"""
        return self.__str__()

    def __str__(self) -> str:
        """Program representation"""
        return f"<Program {self.executable.exec_file.name} ({self.arch.__name__})>"

    @staticmethod
    def from_binary(
        exec_path: Union[pathlib.Path, str],
        output_file: Optional[Union[pathlib.Path, str]] = None,
        database_file: Optional[Union[pathlib.Path, str]] = None,
        debug: bool = False,
        timeout: Optional[int] = 600,
    ) -> Optional[Program]:
        """Generate an export file directly from the binary.

        This methods will export `exec_path` directly using Quokka IDA's plugin if
        installed.

        Arguments:
            exec_path: Binary to export.
            output_file: Where to store the result (by default: near the executable)
            database_file: Where to store IDA database (by default: near the executable)
            timeout: How long should we wait for the export to finish (default: 10 min)
            debug: Activate the debug output

        Returns:
            A |`Program` instance or None if

        Raises:
            FileNotFoundError: If the executable is not found
        """

        exec_path = pathlib.Path(exec_path)
        if not exec_path.is_file():
            raise FileNotFoundError("Missing exec file")

        if output_file is None:
            output_file = exec_path.parent / f"{exec_path.name}.Quokka"
        else:
            output_file = pathlib.Path(output_file)

        if output_file.is_file():
            return Program(output_file, exec_path)

        exec_file = exec_path
        if database_file is None:
            database_file = exec_file.parent / f"{exec_file.name}.i64"
        else:
            database_file = pathlib.Path(database_file)

        additional_options = []
        if not database_file.is_file():
            additional_options.append(f'-o{database_file.with_suffix("")}')
        else:
            exec_file = database_file

        ida_path = os.environ.get("IDA_PATH", "idat64")
        try:
            cmd = (
                [
                    ida_path,
                    "-OQuokkaAuto:true",
                    f"-OQuokkaFile:{output_file}",
                ]
                + additional_options
                + ["-A", f"{exec_file!s}"]
            )

            Program.logger.info("%s", " ".join(cmd))
            result = subprocess.run(
                cmd,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
                env={
                    "TVHEADLESS": "1",
                    "HOME": os.environ["HOME"],
                    "PATH": os.environ.get("PATH", ""),
                    "TERM": "xterm",  # problem with libcurses
                },
                timeout=timeout,
                check=True,
            )
            if debug or result.returncode != 0:
                Program.logger.debug(result.stderr)

        except subprocess.CalledProcessError:
            return None

        if not output_file.is_file():
            return None

        return Program(output_file, exec_path)



"""Methods to use and deal with blocks in a binary."""

#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
import logging
import collections
from functools import cached_property
from typing import TYPE_CHECKING

import quokka
from quokka.types import (
    AddressT,
    BlockType,
    DataType,
    Dict,
    Index,
    Iterator,
    List,
    MutableMapping,
    ReferenceType,
    Set,
)

if TYPE_CHECKING:
    import pypcode
    from quokka.instruction import Instruction
    from typing import Iterator

logger: logging.Logger = logging.getLogger(__name__)


class Block(MutableMapping):
    """Basic Block class

    A basic block is a sequence of instructions without any (basic) incoming flows
    disrupting it (except calls returns).

    While blocks may be serialized in the exported file, a new instance of this class is
    created for each block in the program (so they all have an unique address).

    Arguments:
        block_idx: Index in the protobuf file of the block
        start_address: Starting address of the block
        chunk: Parent chunk (e.g. function) of the block.

    Attributes:
        proto_index: Index inside the protobuf
        parent: A reference to the parent Chunk
        program: A reference to the parent Program
        start: Start address
        fake: Is it a fake block (e.g. belongs to a fake chunk)
        type: Block type
        address_to_index: A mapping of addresses to instruction indexes
        end: End address
        comments: List of comments attached to the block
        references: References mapping attached to the block (TODO(dm): remove me?)
    """

    def __init__(
        self,
        block_idx: Index,
        start_address: AddressT,
        chunk: quokka.Chunk,
    ):
        """Constructor"""
        self.proto_index: Index = block_idx
        self.parent: quokka.Chunk = chunk
        self.program: quokka.Program = chunk.program

        block: "quokka.pb.Quokka.FunctionChunk.Block"
        block = self.program.proto.function_chunks[chunk.proto_index].blocks[block_idx]

        self.start: int = start_address
        self.fake: bool = block.is_fake
        self.type: BlockType = BlockType.from_proto(block.block_type)

        self.address_to_index: Dict[AddressT, Index] = {}
        self._raw_dict: Dict[AddressT, Index] = {}

        current_address: AddressT = self.start
        for instruction_index, instruction_proto_index in enumerate(
            block.instructions_index
        ):
            self.address_to_index[current_address] = instruction_index
            self._raw_dict[current_address] = instruction_proto_index
            current_address += self.program.proto.instructions[
                instruction_proto_index
            ].size

        self.end: int = current_address

        self.comments: Dict[AddressT, str] = {}
        self.references: Dict[str, List[int]] = {"src": [], "dst": []}

    def __setitem__(self, k: AddressT, v: Index) -> None:
        """Update the instructions mapping"""
        self._raw_dict.__setitem__(k, v)

    def __delitem__(self, v: AddressT) -> None:
        """Remove an instruction from the mapping"""
        self._raw_dict.__delitem__(v)

    def add_comment(self, addr: AddressT, value: str) -> None:
        """Set the comment at `addr`.

        Arguments:
            addr: Comment address
            value: Comment value
        """
        self.comments[addr] = value

    @cached_property
    def strings(self) -> List[str]:
        """Compute the list of strings used in this block."""

        strings: Set[str] = set()

        for reference in self.program.references.resolve_block_references(
            self.parent.proto_index,
            self.proto_index,
            ReferenceType.DATA,
            towards=True,
        ):
            reference_source = reference.source
            if (
                isinstance(reference_source, quokka.data.Data)
                and reference_source.type == DataType.ASCII
            ):
                strings.add(reference_source.value)

        return list(strings)

    def __getitem__(self, address: AddressT) -> quokka.Instruction:
        """Retrieve an instruction at `address`."""
        item = self._raw_dict.__getitem__(address)
        return quokka.Instruction(
            proto_index=item,
            inst_index=self.address_to_index[address],
            address=address,
            block=self,
        )

    def __len__(self) -> int:
        """Number of instruction in the block"""
        return len(self._raw_dict)

    def __iter__(self) -> Iterator:
        """Return an iterator over the instruction list"""
        return iter(self._raw_dict)

    @property
    def data_references(self):
        """Return (and compute if needed) the data referenced by this block."""
        data_references: List[quokka.Data] = []
        for instruction in self.values():
            data_references.extend(instruction.data_references)

        return data_references

    @property
    def size(self) -> int:
        """Size of the block.

        This number is the number of instruction * the size of an instruction for
        architecture with fixed length instructions (e.g. ARM).
        """
        return self.end - self.start

    @cached_property
    def constants(self) -> List[int]:
        """Constants used by the block"""
        constants: List[int] = []
        for instruction in self.values():
            constants.extend(instruction.constants)

        return constants

    @property
    def instructions(self) -> Iterator[Instruction]:
        """Accessor of the block instructions"""
        return iter(self.values())

    def __repr__(self) -> str:
        """Block Representation"""
        return (
            f"<Block at 0x{self.start:x} ({self.type}) with {len(self)} instructions>"
        )

    def __hash__(self) -> int:
        """Hash of the block.

        The proto index is guaranteed to be unique so we can use it as an hash and
        forget about un-hashable types.

        TODO(dm):
            Check this
        """
        return self.proto_index

    def successors(self) -> Iterator[AddressT]:
        """(Addresses of the) Successors of the current block."""
        return self.parent.graph.successors(self.start)

    def predecessors(self) -> Iterator[AddressT]:
        """(Addresses of) Predecessors of the current block"""
        return self.parent.graph.predecessors(self.start)

    @property
    def last_instruction(self) -> quokka.Instruction:
        """Direct accessor of the last instruction in the block"""
        deque = collections.deque(self.instructions, maxlen=1)
        return deque.pop()

    @cached_property
    def bytes(self) -> bytes:
        """Retrieve the block bytes

        All bytes for the block are read at once in the file but the result is not
        cached.
        """
        try:
            file_offset: int = self.program.addresser.file(self.start)
        except quokka.NotInFileError:
            logger.warning("Trying to get the bytes for a block not in file.")
            return b""

        # Read all block at once
        block_bytes = self.program.executable.read_bytes(
            offset=file_offset,
            size=self.size,
        )

        return block_bytes

    @cached_property
    def pcode_insts(self) -> List[pypcode.PcodeOp]:
        """Generate PCode instructions for the block

        This method will call the backend Pypcode and generate the instruction for the
        whole block, updating all the instruction inside the block as well.

        However, all instructions will from now be attached to the block itself, and not
        the instructions so the list may differ after some optimizations (e.g.
        len(self.pcode_insts) != sum(len(inst.pcode_insts) for inst in block.values()) )

        Returns:
            A list of PCode instructions

        """
        from quokka.backends.pypcode import pypcode_decode_block

        return pypcode_decode_block(self)



"""Capstone integration"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
import capstone

import quokka
import quokka.analysis

from quokka.types import AddressT, Endianness, Type, Optional


def get_capstone_context(
        arch: Type[quokka.analysis.QuokkaArch],
        endian: Type[Endianness] = Endianness.LITTLE_ENDIAN) -> capstone.Cs:
    """Compute the capstone context for the program

    The Capstone context is used to decode instructions afterwards. Since we are
    interested in most of the details, we already set the details to True.

    Arguments:
        arch: Quokka program architecture

    Returns:
        A capstone Cs instance
    """
    endian_mapping = {
        Endianness.BIG_ENDIAN: capstone.CS_MODE_BIG_ENDIAN,
        Endianness.LITTLE_ENDIAN: capstone.CS_MODE_LITTLE_ENDIAN,
    }
    capstone_endian = endian_mapping[endian]

    mapping = {
        quokka.analysis.ArchARM: (capstone.CS_ARCH_ARM, capstone.CS_MODE_ARM),
        quokka.analysis.ArchARM64: (capstone.CS_ARCH_ARM64, capstone.CS_MODE_ARM),
        quokka.analysis.ArchX86: (capstone.CS_ARCH_X86, capstone.CS_MODE_32),
        quokka.analysis.ArchX64: (capstone.CS_ARCH_X86, capstone.CS_MODE_64),
        quokka.analysis.ArchARMThumb: (
            capstone.CS_ARCH_ARM,
            capstone.CS_MODE_THUMB,
        ),
        quokka.analysis.ArchMIPS: (
            capstone.CS_ARCH_MIPS,
            capstone.CS_MODE_32 + capstone_endian),
        quokka.analysis.ArchMIPS64: (
            capstone.CS_ARCH_MIPS,
            capstone.CS_MODE_64 + capstone_endian),
        quokka.analysis.ArchPPC: (
            capstone.CS_ARCH_PPC,
            capstone.CS_MODE_32 + capstone_endian),
        quokka.analysis.ArchPPC64: (
            capstone.CS_ARCH_PPC,
            capstone.CS_MODE_64 + capstone_endian),
    }

    try:
        capstone_arch, capstone_mode = mapping.get(arch)
    except TypeError as exc:
        raise quokka.CapstoneError("Unable to find the Architecture") from exc

    context = capstone.Cs(capstone_arch, capstone_mode)
    context.detail = True

    return context


def _decode(
    context: capstone.Cs, opcode: bytes, address: AddressT, count: int = 1
) -> Optional[capstone.CsInsn]:
    """Inner method to decode with capstone

    Arguments:
        context: A capstone context (see `get_capstone_context`)
        opcode: Bytes to decode
        address: Address of the instruction
        count: Number of instructions to decode

    Returns:
        A capstone instruction if any are found
    """
    capstone_insts = context.disasm(opcode, address, count)
    return next(capstone_insts, None)


def update_capstone_context(program: quokka.Program, is_thumb: bool) -> capstone.Cs:
    """Returns an appropriate context for Capstone instructions

    For ARM architecture, if the instruction is Thumb, we must use a different context.

    Arguments:
        program: Program to consider
        is_thumb: Is the instruction a thumb one?

    Returns:
        The correct capstone context
    """
    if (
        program.arch
        in (
            quokka.analysis.ArchARM,
            quokka.analysis.ArchARM64,
            quokka.analysis.ArchARMThumb,
        )
        and is_thumb
    ):
        return get_capstone_context(quokka.analysis.ArchARMThumb)

    return program.capstone


def capstone_decode_instruction(
    inst: quokka.Instruction,
) -> Optional[capstone.CsInsn]:
    """Decode an instruction with capstone

    Decode an instruction and retry for ARM to check if the Thumb mode was activated
    The decoding logic is done by the inner method `_decode`.

    Arguments:
        inst: Instruction to translate

    Returns:
        A capstone instruction if it has been decoded
    """

    context: capstone.Cs = update_capstone_context(inst.program, inst.thumb)
    capstone_inst = _decode(context, inst.bytes, inst.address, count=1)

    if capstone_inst is None and context.arch == capstone.CS_ARCH_ARM:
        if context.mode == capstone.CS_MODE_THUMB:
            new_context = get_capstone_context(quokka.analysis.ArchARM)
        else:
            new_context = get_capstone_context(quokka.analysis.ArchARMThumb)

        capstone_inst = _decode(new_context, inst.bytes, inst.address, count=1)

    return capstone_inst



"""Exceptions for quokka.

All exceptions must derive from the QuokkaError.
"""

#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


class QuokkaError(Exception):
    """Base exception in Quokka"""


class NotInFileError(QuokkaError):
    """Raised when trying to read a string outside the file"""


class ChunkMissingError(QuokkaError):
    """Raised when a chunk has not been found"""


class FunctionMissingError(QuokkaError):
    """Raised when a function has not been found"""


class ThunkMissingError(QuokkaError):
    """Raised when a thunk has not been found"""


class InstructionError(QuokkaError):
    """Raised when serious errors in Instructions handling"""


class CapstoneError(QuokkaError):
    """Exceptions used for Capstone integration"""


class PypcodeError(QuokkaError):
    """Main exception for pypcode integration"""



"""Environment module"""
#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
import enum

from quokka.analysis import ArchX64, ArchX86, ArchARM, ArchARM64
from quokka.types import Type

import quokka
import quokka.analysis.calling_convention as cc


def get_calling_convention_for_arch_platform(
    platform: quokka.analysis.Platform,
    arch: Type[quokka.analysis.QuokkaArch],
) -> Type[cc.CallingConvention]:
    """Retrieve the calling convention used for this couple platform/arch

    Arguments:
        platform: Used platform
        arch: Used architecture

    Returns:
        A calling convention
    """
    mapping = {
        quokka.analysis.Platform.LINUX: {
            ArchX64: cc.SystemVAMD,  # Must be before x86
            ArchX86: cc.CCdecl,
            ArchARM: cc.ARMCC,
            ArchARM64: cc.ARM64CC,
        },
        quokka.analysis.Platform.WINDOWS: {
            ArchX86: cc.Stdcall,
            ArchX64: cc.MicrosoftAMD64,
        },
    }

    platform_mapping = mapping.get(platform)
    if platform_mapping is None:
        return cc.CallingConvention

    for architecture, convention in platform_mapping.items():
        if issubclass(arch, architecture):
            return convention

    return cc.CallingConvention


class Environment:
    """Environment base class

    Args:
        platform: Platform
        arch: Architecture

    Attributes:
        platform: Platform
        arch: Architecture
        calling_convention: Calling convention

    """

    def __init__(
        self,
        platform: quokka.analysis.Platform,
        arch: Type[quokka.analysis.arch.QuokkaArch],
    ):
        """Constructor"""
        self.platform: quokka.analysis.Platform = platform
        self.arch: Type[quokka.analysis.arch.QuokkaArch] = arch
        self.calling_convention: Type[
            cc.CallingConvention
        ] = get_calling_convention_for_arch_platform(platform, arch)


class Platform(enum.Enum):
    """Platform enumeration"""

    UNKNOWN = enum.auto()
    WINDOWS = enum.auto()
    LINUX = enum.auto()
    APPLE = enum.auto()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: quokka.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cquokka.proto\x12\x06quokka\"\xfb-\n\x06Quokka\x12!\n\x04meta\x18\x01 \x01(\x0b\x32\x13.quokka.Quokka.Meta\x12\x32\n\rexporter_meta\x18\x02 \x01(\x0b\x32\x1b.quokka.Quokka.ExporterMeta\x12%\n\x06layout\x18\x03 \x03(\x0b\x32\x15.quokka.Quokka.Layout\x12!\n\x04\x64\x61ta\x18\x04 \x03(\x0b\x32\x13.quokka.Quokka.Data\x12)\n\x07structs\x18\x05 \x03(\x0b\x32\x18.quokka.Quokka.Structure\x12\x35\n\x0f\x66unction_chunks\x18\x06 \x03(\x0b\x32\x1c.quokka.Quokka.FunctionChunk\x12\x30\n\x0cinstructions\x18\x07 \x03(\x0b\x32\x1a.quokka.Quokka.Instruction\x12\x11\n\tmnemonics\x18\x08 \x03(\t\x12*\n\tfunctions\x18\t \x03(\x0b\x32\x17.quokka.Quokka.Function\x12,\n\nreferences\x18\n \x03(\x0b\x32\x18.quokka.Quokka.Reference\x12\x14\n\x0cstring_table\x18\x0b \x03(\t\x12\x15\n\roperand_table\x18\x10 \x03(\t\x12(\n\x08operands\x18\x0c \x03(\x0b\x32\x16.quokka.Quokka.Operand\x12\x15\n\rcomment_table\x18\r \x03(\t\x12(\n\x08\x63omments\x18\x0e \x03(\x0b\x32\x16.quokka.Quokka.Comment\x12(\n\x08segments\x18\x0f \x03(\x0b\x32\x16.quokka.Quokka.Segment\x1a\x87\x01\n\x0c\x45xporterMeta\x12.\n\x04mode\x18\x01 \x01(\x0e\x32 .quokka.Quokka.ExporterMeta.Mode\x12\x0f\n\x07version\x18\x02 \x01(\t\"6\n\x04Mode\x12\x0e\n\nMODE_LIGHT\x10\x00\x12\r\n\tMODE_FULL\x10\x01\x12\x0f\n\x0bMODE_NORMAL\x10\x02\x1a\xfe\x06\n\x04Meta\x12\x17\n\x0f\x65xecutable_name\x18\x01 \x01(\t\x12$\n\x03isa\x18\x02 \x01(\x0e\x32\x17.quokka.Quokka.Meta.ISA\x12.\n\x08\x63ompiler\x18\x03 \x01(\x0e\x32\x1c.quokka.Quokka.Meta.Compiler\x12\x41\n\x12\x63\x61lling_convention\x18\x04 \x01(\x0e\x32%.quokka.Quokka.Meta.CallingConvention\x12&\n\x04hash\x18\x05 \x01(\x0b\x32\x18.quokka.Quokka.Meta.Hash\x12\x30\n\tendianess\x18\x06 \x01(\x0e\x32\x1d.quokka.Quokka.Meta.Endianess\x12\x30\n\x0c\x61\x64\x64ress_size\x18\t \x01(\x0e\x32\x1a.quokka.Quokka.AddressSize\x12\x11\n\tbase_addr\x18\x07 \x01(\x04\x12\x13\n\x0bida_version\x18\x08 \x01(\r\x1a\x8a\x01\n\x04Hash\x12\x34\n\thash_type\x18\x01 \x01(\x0e\x32!.quokka.Quokka.Meta.Hash.HashType\x12\x12\n\nhash_value\x18\x02 \x01(\t\"8\n\x08HashType\x12\r\n\tHASH_NONE\x10\x00\x12\x0f\n\x0bHASH_SHA256\x10\x01\x12\x0c\n\x08HASH_MD5\x10\x02\"_\n\x03ISA\x12\x0e\n\nPROC_INTEL\x10\x00\x12\x0c\n\x08PROC_ARM\x10\x01\x12\x0f\n\x0bPROC_DALVIK\x10\x02\x12\x0c\n\x08PROC_PPC\x10\x03\x12\r\n\tPROC_MIPS\x10\x04\x12\x0c\n\x08PROC_UNK\x10\x05\"0\n\tEndianess\x12\n\n\x06\x45ND_LE\x10\x00\x12\n\n\x06\x45ND_BE\x10\x01\x12\x0b\n\x07\x45ND_UNK\x10\x02\"o\n\x08\x43ompiler\x12\x0c\n\x08\x43OMP_GCC\x10\x00\x12\x0b\n\x07\x43OMP_MS\x10\x01\x12\x0b\n\x07\x43OMP_BC\x10\x02\x12\x0f\n\x0b\x43OMP_WATCOM\x10\x03\x12\x0f\n\x0b\x43OMP_VISAGE\x10\x04\x12\x0b\n\x07\x43OMP_BP\x10\x05\x12\x0c\n\x08\x43OMP_UNK\x10\x06\"\x7f\n\x11\x43\x61llingConvention\x12\n\n\x06\x43\x43_UNK\x10\x00\x12\x0c\n\x08\x43\x43_CDECL\x10\x01\x12\x0f\n\x0b\x43\x43_ELLIPSIS\x10\x02\x12\x0e\n\nCC_STDCALL\x10\x03\x12\r\n\tCC_PASCAL\x10\x04\x12\x0f\n\x0b\x43\x43_FASTCALL\x10\x05\x12\x0f\n\x0b\x43\x43_THISCALL\x10\x06\x1a\xa6\x01\n\x07Operand\x12\x0c\n\x04type\x18\x01 \x01(\r\x12\r\n\x05\x66lags\x18\x02 \x01(\r\x12\x12\n\nvalue_type\x18\x03 \x01(\r\x12\x13\n\x0bregister_id\x18\x04 \x01(\r\x12\x11\n\tphrase_id\x18\x05 \x01(\r\x12\r\n\x05value\x18\x06 \x01(\x04\x12\x0f\n\x07\x61\x64\x64ress\x18\x07 \x01(\x03\x12\x0f\n\x07specval\x18\x08 \x01(\x04\x12\x11\n\tspecflags\x18\t \x03(\r\x1au\n\x0bInstruction\x12\x0c\n\x04size\x18\x01 \x01(\r\x12\x16\n\x0emnemonic_index\x18\x02 \x01(\r\x12\x15\n\roperand_index\x18\x03 \x03(\r\x12\x10\n\x08is_thumb\x18\x04 \x01(\x08\x12\x17\n\x0foperand_strings\x18\x05 \x03(\x04\x1a\x87\x04\n\rFunctionChunk\x12\x14\n\x0coffset_start\x18\x01 \x01(\x04\x12\x32\n\x06\x62locks\x18\x02 \x03(\x0b\x32\".quokka.Quokka.FunctionChunk.Block\x12\"\n\x05\x65\x64ges\x18\x03 \x03(\x0b\x32\x13.quokka.Quokka.Edge\x12\x0f\n\x07is_fake\x18\x04 \x01(\x08\x12\x11\n\tis_infile\x18\x05 \x01(\x08\x1a\xe3\x02\n\x05\x42lock\x12\x14\n\x0coffset_start\x18\x01 \x01(\x04\x12\x0f\n\x07is_fake\x18\x02 \x01(\x08\x12\x1a\n\x12instructions_index\x18\x03 \x03(\r\x12@\n\nblock_type\x18\x04 \x01(\x0e\x32,.quokka.Quokka.FunctionChunk.Block.BlockType\"\xd4\x01\n\tBlockType\x12\x15\n\x11\x42LOCK_TYPE_NORMAL\x10\x00\x12\x16\n\x12\x42LOCK_TYPE_INDJUMP\x10\x01\x12\x12\n\x0e\x42LOCK_TYPE_RET\x10\x02\x12\x14\n\x10\x42LOCK_TYPE_NORET\x10\x03\x12\x15\n\x11\x42LOCK_TYPE_CNDRET\x10\x04\x12\x15\n\x11\x42LOCK_TYPE_ENORET\x10\x05\x12\x15\n\x11\x42LOCK_TYPE_EXTERN\x10\x06\x12\x14\n\x10\x42LOCK_TYPE_ERROR\x10\x07\x12\x13\n\x0f\x42LOCK_TYPE_FAKE\x10\x08\x1aV\n\x0f\x42lockIdentifier\x12\x10\n\x08\x62lock_id\x18\x01 \x01(\r\x12\x12\n\x08no_chunk\x18\x02 \x01(\x08H\x00\x12\x12\n\x08\x63hunk_id\x18\x03 \x01(\rH\x00\x42\t\n\x07\x43hunkId\x1a\xf0\x01\n\x04\x45\x64ge\x12/\n\tedge_type\x18\x01 \x01(\x0e\x32\x1c.quokka.Quokka.Edge.EdgeType\x12.\n\x06source\x18\x02 \x01(\x0b\x32\x1e.quokka.Quokka.BlockIdentifier\x12\x33\n\x0b\x64\x65stination\x18\x03 \x01(\x0b\x32\x1e.quokka.Quokka.BlockIdentifier\"R\n\x08\x45\x64geType\x12\x16\n\x12TYPE_UNCONDITIONAL\x10\x00\x12\r\n\tTYPE_TRUE\x10\x01\x12\x0e\n\nTYPE_FALSE\x10\x02\x12\x0f\n\x0bTYPE_SWITCH\x10\x03\x1a\xf6\x04\n\x08\x46unction\x12\x0e\n\x06offset\x18\x01 \x01(\r\x12\x1d\n\x15\x66unction_chunks_index\x18\x02 \x03(\r\x12;\n\rfunction_type\x18\x03 \x01(\x0e\x32$.quokka.Quokka.Function.FunctionType\x12\x0c\n\x04name\x18\x04 \x01(\t\x12(\n\x0b\x63hunk_edges\x18\x05 \x03(\x0b\x32\x13.quokka.Quokka.Edge\x12>\n\x0f\x62lock_positions\x18\x06 \x03(\x0b\x32%.quokka.Quokka.Function.BlockPosition\x12\x14\n\x0cmangled_name\x18\x07 \x01(\t\x1a\x90\x01\n\x08Position\x12\t\n\x01x\x18\x01 \x01(\x05\x12\t\n\x01y\x18\x02 \x01(\x05\x12\x44\n\rposition_type\x18\x03 \x01(\x0e\x32-.quokka.Quokka.Function.Position.PositionType\"(\n\x0cPositionType\x12\n\n\x06\x43\x45NTER\x10\x00\x12\x0c\n\x08TOP_LEFT\x10\x01\x1au\n\rBlockPosition\x12\x30\n\x08\x62lock_id\x18\x01 \x01(\x0b\x32\x1e.quokka.Quokka.BlockIdentifier\x12\x32\n\x08position\x18\x02 \x01(\x0b\x32 .quokka.Quokka.Function.Position\"f\n\x0c\x46unctionType\x12\x0f\n\x0bTYPE_NORMAL\x10\x00\x12\x11\n\rTYPE_IMPORTED\x10\x01\x12\x10\n\x0cTYPE_LIBRARY\x10\x02\x12\x0e\n\nTYPE_THUNK\x10\x03\x12\x10\n\x0cTYPE_INVALID\x10\x04\x1a\xff\x01\n\x06Layout\x12\x39\n\raddress_range\x18\x01 \x01(\x0b\x32\".quokka.Quokka.Layout.AddressRange\x12\x35\n\x0blayout_type\x18\x02 \x01(\x0e\x32 .quokka.Quokka.Layout.LayoutType\x1a\x33\n\x0c\x41\x64\x64ressRange\x12\x15\n\rstart_address\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\x04\"N\n\nLayoutType\x12\x0e\n\nLAYOUT_UNK\x10\x00\x12\x0f\n\x0bLAYOUT_CODE\x10\x01\x12\x0f\n\x0bLAYOUT_DATA\x10\x02\x12\x0e\n\nLAYOUT_GAP\x10\x03\x1a\xae\x01\n\x04\x44\x61ta\x12\x0e\n\x06offset\x18\x01 \x01(\x04\x12%\n\x04type\x18\x02 \x01(\x0e\x32\x17.quokka.Quokka.DataType\x12\x0e\n\x04size\x18\x03 \x01(\rH\x00\x12\x11\n\x07no_size\x18\x04 \x01(\x08H\x00\x12\x13\n\x0bvalue_index\x18\x05 \x01(\r\x12\x12\n\nname_index\x18\x06 \x01(\r\x12\x17\n\x0fnot_initialized\x18\x07 \x01(\x08\x42\n\n\x08\x44\x61taSize\x1a\xe1\x02\n\tStructure\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x34\n\x04type\x18\x02 \x01(\x0e\x32&.quokka.Quokka.Structure.StructureType\x12\x0c\n\x04size\x18\x03 \x01(\r\x12\x15\n\rvariable_size\x18\x04 \x01(\x08\x12\x30\n\x07members\x18\x05 \x03(\x0b\x32\x1f.quokka.Quokka.Structure.Member\x1aj\n\x06Member\x12\x0e\n\x06offset\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12%\n\x04type\x18\x03 \x01(\x0e\x32\x17.quokka.Quokka.DataType\x12\x0c\n\x04size\x18\x04 \x01(\r\x12\r\n\x05value\x18\x05 \x01(\x03\"M\n\rStructureType\x12\x0f\n\x0bTYPE_STRUCT\x10\x00\x12\r\n\tTYPE_ENUM\x10\x01\x12\x0e\n\nTYPE_UNION\x10\x02\x12\x0c\n\x08TYPE_UNK\x10\x03\x1a\xe4\x01\n\x07\x43omment\x12\x30\n\x04type\x18\x01 \x01(\x0e\x32\".quokka.Quokka.Comment.CommentType\x12\x12\n\nstring_idx\x18\x02 \x01(\r\x12)\n\x08location\x18\x03 \x01(\x0b\x32\x17.quokka.Quokka.Location\"h\n\x0b\x43ommentType\x12\x17\n\x13\x43OMMENT_INSTRUCTION\x10\x00\x12\x14\n\x10\x43OMMENT_FUNCTION\x10\x01\x12\x15\n\x11\x43OMMENT_STRUCTURE\x10\x02\x12\x13\n\x0f\x43OMMENT_INVALID\x10\x03\x1a\xc4\x03\n\x08Location\x12\x12\n\x08inst_idx\x18\x01 \x01(\x04H\x00\x12\x12\n\x08\x64\x61ta_idx\x18\x02 \x01(\rH\x00\x12\x44\n\x0fstruct_position\x18\x03 \x01(\x0b\x32).quokka.Quokka.Location.StructurePositionH\x00\x12M\n\x14instruction_position\x18\x04 \x01(\x0b\x32-.quokka.Quokka.Location.InstructionIdentifierH\x00\x12\x16\n\x0c\x66unction_idx\x18\x05 \x01(\rH\x00\x12\x13\n\tchunk_idx\x18\x06 \x01(\rH\x00\x1a\x61\n\x11StructurePosition\x12\x15\n\rstructure_idx\x18\x01 \x01(\r\x12\x14\n\nmember_idx\x18\x02 \x01(\rH\x00\x12\x13\n\tno_member\x18\x03 \x01(\x08H\x00\x42\n\n\x08MemberId\x1a[\n\x15InstructionIdentifier\x12\x16\n\x0e\x66unc_chunk_idx\x18\x01 \x01(\r\x12\x11\n\tblock_idx\x18\x02 \x01(\r\x12\x17\n\x0finstruction_idx\x18\x03 \x01(\rB\x0e\n\x0cLocationType\x1a\xf9\x01\n\tReference\x12\'\n\x06source\x18\x01 \x01(\x0b\x32\x17.quokka.Quokka.Location\x12,\n\x0b\x64\x65stination\x18\x02 \x01(\x0b\x32\x17.quokka.Quokka.Location\x12>\n\x0ereference_type\x18\x03 \x01(\x0e\x32&.quokka.Quokka.Reference.ReferenceType\"U\n\rReferenceType\x12\x0c\n\x08REF_CALL\x10\x00\x12\x0c\n\x08REF_DATA\x10\x01\x12\x0c\n\x08REF_ENUM\x10\x02\x12\r\n\tREF_STRUC\x10\x03\x12\x0b\n\x07REF_UNK\x10\x04\x1a\xdd\x03\n\x07Segment\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\nstart_addr\x18\x02 \x01(\x04\x12\x0c\n\x04size\x18\x03 \x01(\x04\x12\x13\n\x0bpermissions\x18\x04 \x01(\r\x12)\n\x04type\x18\x05 \x01(\x0e\x32\x1b.quokka.Quokka.Segment.Type\x12\x30\n\x0c\x61\x64\x64ress_size\x18\x06 \x01(\x0e\x32\x1a.quokka.Quokka.AddressSize\x12\x15\n\x0b\x66ile_offset\x18\x07 \x01(\x04H\x00\x12\x13\n\tno_offset\x18\x08 \x01(\x08H\x00\"\xa4\x01\n\x04Type\x12\x0f\n\x0bSEGMENT_UNK\x10\x00\x12\x10\n\x0cSEGMENT_CODE\x10\x01\x12\x10\n\x0cSEGMENT_DATA\x10\x02\x12\x0f\n\x0bSEGMENT_BSS\x10\x03\x12\x10\n\x0cSEGMENT_NULL\x10\x04\x12\x12\n\x0eSEGMENT_EXTERN\x10\x05\x12\x12\n\x0eSEGMENT_NORMAL\x10\x06\x12\x1c\n\x18SEGMENT_ABSOLUTE_SYMBOLS\x10\x07\"N\n\x07\x42itness\x12\x0e\n\nBITNESS_64\x10\x00\x12\x0e\n\nBITNESS_32\x10\x01\x12\x0e\n\nBITNESS_16\x10\x02\x12\x13\n\x0f\x42ITNESS_UNKNOWN\x10\x03\x42\r\n\x0boffset_type\"5\n\x0b\x41\x64\x64ressSize\x12\x0c\n\x08\x41\x44\x44R_UNK\x10\x00\x12\x0b\n\x07\x41\x44\x44R_32\x10\x01\x12\x0b\n\x07\x41\x44\x44R_64\x10\x02\"\xbb\x01\n\x08\x44\x61taType\x12\x0c\n\x08TYPE_UNK\x10\x00\x12\n\n\x06TYPE_B\x10\x01\x12\n\n\x06TYPE_W\x10\x02\x12\x0b\n\x07TYPE_DW\x10\x03\x12\x0b\n\x07TYPE_QW\x10\x04\x12\x0b\n\x07TYPE_OW\x10\x05\x12\x0e\n\nTYPE_FLOAT\x10\x06\x12\x0f\n\x0bTYPE_DOUBLE\x10\x07\x12\x0e\n\nTYPE_ASCII\x10\x08\x12\x0f\n\x0bTYPE_STRUCT\x10\t\x12\x0e\n\nTYPE_ALIGN\x10\n\x12\x10\n\x0cTYPE_POINTER\x10\x0b\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'quokka_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_QUOKKA']._serialized_start=25
  _globals['_QUOKKA']._serialized_end=5908
  _globals['_QUOKKA_EXPORTERMETA']._serialized_start=648
  _globals['_QUOKKA_EXPORTERMETA']._serialized_end=783
  _globals['_QUOKKA_EXPORTERMETA_MODE']._serialized_start=729
  _globals['_QUOKKA_EXPORTERMETA_MODE']._serialized_end=783
  _globals['_QUOKKA_META']._serialized_start=786
  _globals['_QUOKKA_META']._serialized_end=1680
  _globals['_QUOKKA_META_HASH']._serialized_start=1153
  _globals['_QUOKKA_META_HASH']._serialized_end=1291
  _globals['_QUOKKA_META_HASH_HASHTYPE']._serialized_start=1235
  _globals['_QUOKKA_META_HASH_HASHTYPE']._serialized_end=1291
  _globals['_QUOKKA_META_ISA']._serialized_start=1293
  _globals['_QUOKKA_META_ISA']._serialized_end=1388
  _globals['_QUOKKA_META_ENDIANESS']._serialized_start=1390
  _globals['_QUOKKA_META_ENDIANESS']._serialized_end=1438
  _globals['_QUOKKA_META_COMPILER']._serialized_start=1440
  _globals['_QUOKKA_META_COMPILER']._serialized_end=1551
  _globals['_QUOKKA_META_CALLINGCONVENTION']._serialized_start=1553
  _globals['_QUOKKA_META_CALLINGCONVENTION']._serialized_end=1680
  _globals['_QUOKKA_OPERAND']._serialized_start=1683
  _globals['_QUOKKA_OPERAND']._serialized_end=1849
  _globals['_QUOKKA_INSTRUCTION']._serialized_start=1851
  _globals['_QUOKKA_INSTRUCTION']._serialized_end=1968
  _globals['_QUOKKA_FUNCTIONCHUNK']._serialized_start=1971
  _globals['_QUOKKA_FUNCTIONCHUNK']._serialized_end=2490
  _globals['_QUOKKA_FUNCTIONCHUNK_BLOCK']._serialized_start=2135
  _globals['_QUOKKA_FUNCTIONCHUNK_BLOCK']._serialized_end=2490
  _globals['_QUOKKA_FUNCTIONCHUNK_BLOCK_BLOCKTYPE']._serialized_start=2278
  _globals['_QUOKKA_FUNCTIONCHUNK_BLOCK_BLOCKTYPE']._serialized_end=2490
  _globals['_QUOKKA_BLOCKIDENTIFIER']._serialized_start=2492
  _globals['_QUOKKA_BLOCKIDENTIFIER']._serialized_end=2578
  _globals['_QUOKKA_EDGE']._serialized_start=2581
  _globals['_QUOKKA_EDGE']._serialized_end=2821
  _globals['_QUOKKA_EDGE_EDGETYPE']._serialized_start=2739
  _globals['_QUOKKA_EDGE_EDGETYPE']._serialized_end=2821
  _globals['_QUOKKA_FUNCTION']._serialized_start=2824
  _globals['_QUOKKA_FUNCTION']._serialized_end=3454
  _globals['_QUOKKA_FUNCTION_POSITION']._serialized_start=3087
  _globals['_QUOKKA_FUNCTION_POSITION']._serialized_end=3231
  _globals['_QUOKKA_FUNCTION_POSITION_POSITIONTYPE']._serialized_start=3191
  _globals['_QUOKKA_FUNCTION_POSITION_POSITIONTYPE']._serialized_end=3231
  _globals['_QUOKKA_FUNCTION_BLOCKPOSITION']._serialized_start=3233
  _globals['_QUOKKA_FUNCTION_BLOCKPOSITION']._serialized_end=3350
  _globals['_QUOKKA_FUNCTION_FUNCTIONTYPE']._serialized_start=3352
  _globals['_QUOKKA_FUNCTION_FUNCTIONTYPE']._serialized_end=3454
  _globals['_QUOKKA_LAYOUT']._serialized_start=3457
  _globals['_QUOKKA_LAYOUT']._serialized_end=3712
  _globals['_QUOKKA_LAYOUT_ADDRESSRANGE']._serialized_start=3581
  _globals['_QUOKKA_LAYOUT_ADDRESSRANGE']._serialized_end=3632
  _globals['_QUOKKA_LAYOUT_LAYOUTTYPE']._serialized_start=3634
  _globals['_QUOKKA_LAYOUT_LAYOUTTYPE']._serialized_end=3712
  _globals['_QUOKKA_DATA']._serialized_start=3715
  _globals['_QUOKKA_DATA']._serialized_end=3889
  _globals['_QUOKKA_STRUCTURE']._serialized_start=3892
  _globals['_QUOKKA_STRUCTURE']._serialized_end=4245
  _globals['_QUOKKA_STRUCTURE_MEMBER']._serialized_start=4060
  _globals['_QUOKKA_STRUCTURE_MEMBER']._serialized_end=4166
  _globals['_QUOKKA_STRUCTURE_STRUCTURETYPE']._serialized_start=4168
  _globals['_QUOKKA_STRUCTURE_STRUCTURETYPE']._serialized_end=4245
  _globals['_QUOKKA_COMMENT']._serialized_start=4248
  _globals['_QUOKKA_COMMENT']._serialized_end=4476
  _globals['_QUOKKA_COMMENT_COMMENTTYPE']._serialized_start=4372
  _globals['_QUOKKA_COMMENT_COMMENTTYPE']._serialized_end=4476
  _globals['_QUOKKA_LOCATION']._serialized_start=4479
  _globals['_QUOKKA_LOCATION']._serialized_end=4931
  _globals['_QUOKKA_LOCATION_STRUCTUREPOSITION']._serialized_start=4725
  _globals['_QUOKKA_LOCATION_STRUCTUREPOSITION']._serialized_end=4822
  _globals['_QUOKKA_LOCATION_INSTRUCTIONIDENTIFIER']._serialized_start=4824
  _globals['_QUOKKA_LOCATION_INSTRUCTIONIDENTIFIER']._serialized_end=4915
  _globals['_QUOKKA_REFERENCE']._serialized_start=4934
  _globals['_QUOKKA_REFERENCE']._serialized_end=5183
  _globals['_QUOKKA_REFERENCE_REFERENCETYPE']._serialized_start=5098
  _globals['_QUOKKA_REFERENCE_REFERENCETYPE']._serialized_end=5183
  _globals['_QUOKKA_SEGMENT']._serialized_start=5186
  _globals['_QUOKKA_SEGMENT']._serialized_end=5663
  _globals['_QUOKKA_SEGMENT_TYPE']._serialized_start=5404
  _globals['_QUOKKA_SEGMENT_TYPE']._serialized_end=5568
  _globals['_QUOKKA_SEGMENT_BITNESS']._serialized_start=5570
  _globals['_QUOKKA_SEGMENT_BITNESS']._serialized_end=5648
  _globals['_QUOKKA_ADDRESSSIZE']._serialized_start=5665
  _globals['_QUOKKA_ADDRESSSIZE']._serialized_end=5718
  _globals['_QUOKKA_DATATYPE']._serialized_start=5721
  _globals['_QUOKKA_DATATYPE']._serialized_end=5908
# @@protoc_insertion_point(module_scope)



#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
import quokka
from quokka.types import List, Dict, Any, RegType


class Replacer:
    """Replacer class

    Warning: This class has been used in some experiments but will/should probably be
    removed because it does not belong to the project.
    """

    ignored_registers: List[str]
    ignored_mnemonics: List[str]

    regs: RegType
    compared_mnemonics: List[str]
    calling_registers: List[RegType]

    replacement: Dict[RegType, RegType]

    @staticmethod
    def norm_mnemonic(mnemonic: str) -> str:
        """
        Norm a mnemonic (remove everything after '.').
        Examples:
            cmp.w -> CMP
            mov -> MOV

        Arguments:
            mnemonic: Mnemonic to norm

        Returns:
            Normed mnemonic
        """
        if "." in mnemonic:
            return mnemonic[: mnemonic.rfind(".")].upper()

        return mnemonic.upper()

    def get(self, item) -> RegType:
        assert self.replacement is not None
        if isinstance(item, RegType):
            return self.replacement.get(item, item)
        return item

    def is_ignored(self, item) -> bool:
        assert self.ignored_registers is not None
        return self.get(item) in self.ignored_registers

    def __call__(self, *args: Any, **kwargs: Any) -> RegType:
        return self.get(*args, **kwargs)

    def is_ignored_mnemonics(self, mnemonic: str) -> bool:
        assert self.ignored_mnemonics is not None
        return self.norm_mnemonic(mnemonic) in self.ignored_mnemonics

    def is_compared_mnemonic(self, mnemonic: str) -> bool:
        assert self.compared_mnemonics is not None
        return self.norm_mnemonic(mnemonic) in self.compared_mnemonics

    def calling_convention(self) -> List[RegType]:
        assert self.calling_registers is not None
        return self.calling_registers

    def argument_registers(self, platform: quokka.analysis.Platform) -> List:
        return []



"""Quokka : A pythonic program manipulation tool

Quokka is a module designed to help using an exported program from a Python(ic) API.
It must be used with the exported version of a file (a Quokka file) and the binary
itself.
"""

#  Copyright 2022-2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from quokka.version import __version__

import quokka.analysis
import quokka.backends
import quokka.quokka_pb2 as pb

from quokka.addresser import Addresser

from quokka.block import Block

from quokka.data import Data, DataHolder

from quokka.exc import (
    QuokkaError,
    NotInFileError,
    ChunkMissingError,
    FunctionMissingError,
    ThunkMissingError,
    InstructionError,
    CapstoneError,
    PypcodeError,
)

from quokka.executable import Executable

from quokka.function import (
    dereference_thunk,
    get_degrees,
    Chunk,
    SuperChunk,
    Function,
)

from quokka.instruction import Operand, Instruction

from quokka.program import Program

from quokka.reference import Reference, ReferencesLocation, References

from quokka.segment import Segment

from quokka.structure import Structure, StructureMember

from quokka.utils import (
    md5_file,
    sha256_file,
    check_hash,
    get_isa,
    convert_address_size,
    get_arch,
    parse_version,
)

__all__ = [
    # From addresser.py
    "Addresser",
    # From block.py
    "Block",
    # From data.py
    "Data",
    "DataHolder",
    # From exc.py
    "QuokkaError",
    "NotInFileError",
    "ChunkMissingError",
    "FunctionMissingError",
    "ThunkMissingError",
    "InstructionError",
    "CapstoneError",
    "PypcodeError",
    # From executable.py
    "Executable",
    # From functions.py
    "dereference_thunk",
    "get_degrees",
    "Chunk",
    "SuperChunk",
    "Function",
    # From instructions;py
    "Operand",
    "Instruction",
    # From program.py
    "Program",
    # From reference.py
    "Reference",
    "References",
    "ReferencesLocation",
    # From segment.py
    "Segment",
    # From structure.py
    "Structure",
    "StructureMember",
    # From utils.py
    "md5_file",
    "sha256_file",
    "check_hash",
    "get_isa",
    "convert_address_size",
    "get_arch",
]



#!/usr/bin/env python3
# coding: utf-8

import logging
import os.path
from pathlib import Path
from typing import Generator

import magic
import click
import subprocess
from multiprocessing import Pool, Queue, Manager
import queue

BINARY_FORMAT = {
    "application/x-dosexec",
    "application/x-sharedlib",
    "application/x-mach-binary",
    "application/x-executable",
    "application/x-pie-executable",
}

EXTENSIONS_WHITELIST = {"application/octet-stream": [".dex"]}

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"], max_content_width=300)

class Bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def recursive_file_iter(p: Path) -> Generator[Path, None, None]:
    if p.is_file():
        mime_type = magic.from_file(p, mime=True)
        if mime_type not in BINARY_FORMAT and p.suffix not in EXTENSIONS_WHITELIST.get(
            mime_type, []
        ):
            pass
        else:
            yield p
    elif p.is_dir():
        for f in p.iterdir():
            yield from recursive_file_iter(f)


def do_quokka(exec_path: Path) -> bool:
    output_file = Path(str(exec_path)+".quokka")

    if output_file.is_file():
        return True

    database_file = exec_path.parent / f"{exec_path.name}.i64"

    additional_options = []
    if not database_file.is_file():
        additional_options.append(f'-o{database_file.with_suffix("")}')
    else:
        exec_path = database_file

    ida_path = os.environ.get("IDA_PATH", "idat64")
    try:
        cmd = [ida_path, "-OQuokkaAuto:true", f"-OQuokkaFile:{output_file}"] + additional_options + ["-A", f"{exec_path!s}"]

        result = subprocess.run(
            cmd,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            env={
                "TVHEADLESS": "1",
                "HOME": os.environ["HOME"],
                "PATH": os.environ.get("PATH", ""),
                "TERM": "xterm",  # problem with libcurses
            },
            check=True,
        )
    except subprocess.CalledProcessError:
        return False

    if output_file.is_file():
        return True
    else:
        return False

def export_job(ingress, egress) -> bool:
    while True:
        try:
            file = ingress.get(timeout=0.5)
            res = do_quokka(file)
            egress.put((file, res))
        except queue.Empty:
            pass
        except KeyboardInterrupt:
            break


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-i",
    "--ida-path",
    type=click.Path(exists=True),
    default=None,
    help="IDA Pro installation directory",
)
@click.option("-t", "--threads", type=int, default=1, help="Thread number to use")
@click.option("-v", "--verbose", count=True, help="To activate or not the verbosity")
@click.argument("input_file", type=click.Path(exists=True), metavar="<binary file|directory>")
def main(ida_path: str, input_file: str, threads: int, verbose: bool) -> None:
    """
    quokka-cli is a very simple utility to generate a .Quokka file
    for a given binary or a directory. It all open the binary file and export files
    seamlessly.

    :param ida_path: Path to the IDA Pro installation directory
    :param input_file: Path of the binary to export
    :param threads: number of threads to use
    :param verbose: To activate or not the verbosity
    :return: None
    """

    logging.basicConfig(
        format="%(message)s", level=logging.DEBUG if verbose else logging.INFO
    )

    if ida_path:
        os.environ["IDA_PATH"] = Path(ida_path).absolute().as_posix()

    root_path = Path(input_file)

    manager = Manager()
    ingress = manager.Queue()
    egress = manager.Queue()
    pool = Pool(threads)

    # Launch all workers
    for _ in range(threads):
        pool.apply_async(export_job, (ingress, egress))

    # Pre-fill ingress queue
    total = 0
    for file in recursive_file_iter(root_path):
        ingress.put(file)
        total += 1

    logging.info(f"Start exporting {total} binaries")

    i = 0
    while True:
        item = egress.get()
        i += 1
        path, res = item
        if res:
            pp_res = Bcolors.OKGREEN + "OK" + Bcolors.ENDC
        else:
            pp_res = Bcolors.FAIL + "KO" + Bcolors.ENDC
        logging.info(f"[{i}/{total}] {str(path) + '.quokka'} [{pp_res}]")
        if i == total:
            break

    pool.terminate()



if __name__ == "__main__":
    main()
